import os
import shlex

from jobbergate_cli.subapps.applications.application_helpers import get_file_list, get_running_jobs
from jobbergate_cli.text_tools import dedent_all


def test_get_running_jobs(mocker):
    """
    Test that the ``get_running_jobs()`` funciton gets a list of running jobs (and the users) from squeue.

    Assert that the ``user_only`` argument is handled correctly. Also verify that if an exception is thrown
    or if the squeue command fails an empty list is returned.
    """
    mocker.patch(
        "jobbergate_cli.subapps.applications.application_helpers.getuser",
        return_value="dummy-user",
    )
    mocked_subprocess = mocker.patch("jobbergate_cli.subapps.applications.application_helpers.subprocess")
    mocked_run = mocked_subprocess.run

    mocked_output = mocker.MagicMock()
    mocked_output.stdout = dedent_all(
        """
              77 foo
              88 bar
              99 "baz bazzington"
        """,
        "",  # For the extra newline!
    ).encode("utf-8")

    mocked_run.return_value = mocked_output

    assert get_running_jobs() == [
        "77 foo",
        "88 bar",
        "99 baz bazzington",
    ]
    mocked_run.assert_called_once_with(
        shlex.split("""squeue --format="%.8A %j" --noheader --user=dummy-user"""),
        capture_output=True,
        check=True,
    )

    mocked_run.reset_mock()

    # squeue found no jobs
    mocked_output.stdout = "".encode("utf-8")
    mocked_run.return_value = mocked_output

    assert get_running_jobs() == []
    mocked_run.assert_called_once_with(
        shlex.split("""squeue --format="%.8A %j" --noheader --user=dummy-user"""),
        capture_output=True,
        check=True,
    )

    mocked_run.reset_mock()

    get_running_jobs(user_only=False)
    mocked_run.assert_called_once_with(
        shlex.split("""squeue --format="%.8A %j" --noheader"""),
        capture_output=True,
        check=True,
    )

    mocked_run.reset_mock()
    mocked_run.side_effect = RuntimeError("BOOM!")

    assert get_running_jobs() == []


def test_get_file_list__basic(tmp_path, temp_cd):
    """
    Test that the ``get_file_list()`` method returns a list of files in a path matching some search term.

    Assert that the function works with no arguments, with only the target path, and with the search term.
    Also assert that case is ignored.
    """
    file1 = tmp_path / "file1.py"
    file1.write_text("foo")
    os.utime(file1, (0, 2))

    file2 = tmp_path / "FILE2.txt"
    file2.write_text("bar")
    os.utime(file2, (0, 3))

    dir1 = tmp_path / "dir1"
    dir1.mkdir()

    file3 = dir1 / "file3.py"
    file3.write_text("baz")
    os.utime(file3, (0, 1))

    with temp_cd(tmp_path):
        assert get_file_list() == [
            "FILE2.txt",
            "file1.py",
        ]

    assert get_file_list(tmp_path) == [
        "FILE2.txt",
        "file1.py",
    ]

    assert get_file_list(tmp_path, search_term="*") == [
        "FILE2.txt",
        "file1.py",
    ]

    assert get_file_list(tmp_path, search_term="*.py") == [
        "file1.py",
    ]

    assert get_file_list(tmp_path, search_term="*FILE*") == [
        "FILE2.txt",
        "file1.py",
    ]



import datetime
import importlib
import pathlib
from unittest.mock import MagicMock

import httpx
import pytest
import respx

from jobbergate_cli.constants import (
    JOBBERGATE_APPLICATION_CONFIG,
    JOBBERGATE_APPLICATION_CONFIG_FILE_NAME,
    JOBBERGATE_APPLICATION_MODULE_FILE_NAME,
    FileType,
)
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.schemas import (
    ApplicationResponse,
    JobbergateApplicationConfig,
    LocalApplication,
    LocalTemplateFile,
    LocalWorkflowFile,
    TemplateFileResponse,
    WorkflowFileResponse,
)
from jobbergate_cli.subapps.applications.application_base import JobbergateApplicationBase
from jobbergate_cli.subapps.applications.questions import Text
from jobbergate_cli.subapps.applications.tools import (
    ApplicationRuntime,
    fetch_application_data,
    fetch_application_data_locally,
    load_application_config_from_source,
    load_default_config,
    save_application_files,
    upload_application,
)


def test_load_default_config():
    default_config = load_default_config()
    assert default_config == JOBBERGATE_APPLICATION_CONFIG
    default_config["foo"] = "bar"
    assert default_config != JOBBERGATE_APPLICATION_CONFIG


def test_fetch_application_data__success__using_id(
    respx_mock,
    dummy_context,
    dummy_application_data,
    dummy_domain,
):
    app_data = dummy_application_data[0]
    app_id = app_data["id"]
    fetch_route = respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates/{app_id}")
    fetch_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=app_data,
        ),
    )

    result = fetch_application_data(dummy_context, id_or_identifier=app_id)
    assert fetch_route.called
    assert result == ApplicationResponse.model_validate(app_data)


def test_fetch_application_data__success__using_identifier(
    respx_mock,
    dummy_context,
    dummy_application_data,
    dummy_domain,
):
    app_data = dummy_application_data[0]
    app_identifier = app_data["identifier"]
    fetch_route = respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates/{app_identifier}")
    fetch_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=app_data,
        ),
    )

    result = fetch_application_data(dummy_context, id_or_identifier=app_identifier)
    assert fetch_route.called
    assert result == ApplicationResponse.model_validate(app_data)


def test_fetch_application_data_locally__success(
    dummy_module_source, dummy_config_source, dummy_template_source, tmp_path
):
    expected_config = load_application_config_from_source(dummy_config_source)

    config_file_path = tmp_path / "jobbergate.yaml"
    config_file_path.write_text(dummy_config_source)

    module_file_path = tmp_path / "jobbergate.py"
    module_file_path.write_text(dummy_module_source)

    template_file_path = tmp_path / "test-job-script.py.j2"
    template_file_path.write_text(dummy_template_source)

    result = fetch_application_data_locally(tmp_path)

    assert result == LocalApplication(
        template_vars=expected_config.application_config,
        template_files=[
            LocalTemplateFile(filename="test-job-script.py.j2", path=template_file_path, file_type=FileType.ENTRYPOINT)
        ],
        workflow_files=[
            LocalWorkflowFile(
                filename="jobbergate.py",
                path=module_file_path,
                runtime_config=expected_config.jobbergate_config.model_dump(),
            )
        ],
    )


def test_fetch_application_data_locally__fails_if_application_path_doesnt_exist():
    non_exist = pathlib.Path("/does/not/exist")
    with pytest.raises(Abort, match=f"Application directory {non_exist} does not exist"):
        fetch_application_data_locally(non_exist)


def test_fetch_application_data_locally_fails_if_config_file_not_found(tmp_path):
    module_file_path = tmp_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
    module_file_path.write_text("")

    template_file_path = tmp_path / "test-job-script.py.j2"
    template_file_path.write_text("")

    with pytest.raises(
        Abort, match=f"Application config file {JOBBERGATE_APPLICATION_CONFIG_FILE_NAME} does not exist"
    ):
        fetch_application_data_locally(tmp_path)


def test_fetch_application_data_locally_fails_if_module_file_not_found(tmp_path):
    config_file_path = tmp_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
    config_file_path.write_text("")

    template_file_path = tmp_path / "test-job-script.py.j2"
    template_file_path.write_text("")

    with pytest.raises(
        Abort, match=f"Application module file {JOBBERGATE_APPLICATION_MODULE_FILE_NAME} does not exist"
    ):
        fetch_application_data_locally(tmp_path)


def test_fetch_application_data_locally_fails_if_no_template_files_found(tmp_path):
    config_file_path = tmp_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
    config_file_path.write_text("")

    module_file_path = tmp_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
    module_file_path.write_text("")

    with pytest.raises(Abort, match="No template files found in"):
        fetch_application_data_locally(tmp_path)


class TestUploadApplicationFiles:
    """Test the upload_application_files function."""

    application_id = 1

    @pytest.fixture(scope="function")
    def mocked_routes(self, dummy_domain):
        app_mock = respx.mock(base_url=dummy_domain)

        app_mock.put(
            path=f"/jobbergate/job-script-templates/{self.application_id}",
            name="update",
        ).respond(httpx.codes.OK)

        app_mock.put(
            path=f"/jobbergate/job-script-templates/{self.application_id}/upload/template/ENTRYPOINT",
            name="upload_template",
        ).respond(httpx.codes.OK)

        app_mock.put(
            path=f"/jobbergate/job-script-templates/{self.application_id}/upload/workflow",
            name="upload_workflow",
        ).respond(httpx.codes.OK)

        yield app_mock

    @respx.mock(assert_all_called=True, assert_all_mocked=True)
    def test_upload_application__success(
        self,
        dummy_application_dir,
        dummy_context,
        mocked_routes,
    ):
        with mocked_routes as routes:
            upload_application(dummy_context, dummy_application_dir, self.application_id)

            # Ensure just the filename is included, nothing extra from path
            filename_check = b'filename="jobbergate.py"\r\n'
            assert routes["upload_workflow"].call_count == 1
            assert filename_check in routes["upload_workflow"].calls[0].request.content

            filename_check = b'filename="job-script-template.py.j2"\r\n'
            assert routes["upload_template"].call_count == 1
            assert filename_check in routes["upload_template"].calls[0].request.content

    @respx.mock(assert_all_mocked=True)
    def test_upload_application__fails_directory_does_not_exists(self, dummy_application_dir, dummy_context):
        application_path = dummy_application_dir / "does-not-exist"
        with pytest.raises(Abort, match="Application directory"):
            upload_application(dummy_context, application_path, self.application_id)

    @respx.mock(assert_all_mocked=True)
    def test_upload_application__fails_config_file_not_found(self, dummy_application_dir, dummy_context):
        file_path = dummy_application_dir / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
        file_path.unlink()
        with pytest.raises(Abort, match="Application config file"):
            upload_application(dummy_context, dummy_application_dir, self.application_id)

    @respx.mock(assert_all_mocked=True)
    def test_upload_application__fails_module_file_not_found(self, dummy_application_dir, dummy_context):
        file_path = dummy_application_dir / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
        file_path.unlink()
        with pytest.raises(Abort, match="Application module file"):
            upload_application(dummy_context, dummy_application_dir, self.application_id)

    @respx.mock(assert_all_mocked=True)
    def test_upload_application__fails_no_template_found(self, dummy_application_dir, dummy_context):
        file_path = dummy_application_dir / "templates" / "job-script-template.py.j2"
        file_path.unlink()
        with pytest.raises(Abort, match="No template files found in"):
            upload_application(dummy_context, dummy_application_dir, self.application_id)


class TestDownloadApplicationFiles:
    """
    Test the save_application_files function.
    """

    def test_save_application_files__no_files(self, tmp_path, dummy_context):
        """
        Test that the function returns an empty list if there are no files to download.
        """
        application_data = ApplicationResponse(
            id=1,
            name="dummy",
            owner_email="dummy@email.com",
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
        )

        desired_list_of_files = []

        assert save_application_files(dummy_context, application_data, tmp_path) == desired_list_of_files
        assert list(tmp_path.rglob("*")) == desired_list_of_files

    def test_save_application_files__all_files(
        self,
        respx_mock,
        dummy_domain,
        dummy_context,
        tmp_path,
        dummy_module_source,
        dummy_config_source,
        dummy_template_source,
    ):
        """
        Test that the function downloads all files.
        """

        application_config = load_application_config_from_source(dummy_config_source)

        application_data = ApplicationResponse(
            id=1,
            name="dummy",
            owner_email="dummy@email.com",
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
            template_vars=application_config.application_config,
            template_files=[
                TemplateFileResponse(
                    filename="test-job-script.py.j2",
                    parent_id=1,
                    created_at=datetime.datetime.now(),
                    updated_at=datetime.datetime.now(),
                    file_type="ENTRYPOINT",
                )
            ],
            workflow_files=[
                WorkflowFileResponse(
                    filename="jobbergate.py",
                    parent_id=1,
                    runtime_config=application_config.jobbergate_config.model_dump(),
                    created_at=datetime.datetime.now(),
                    updated_at=datetime.datetime.now(),
                )
            ],
        )

        get_template_routes = [respx_mock.get(f"{dummy_domain}{t.path}") for t in application_data.template_files]
        for route in get_template_routes:
            route.mock(
                return_value=httpx.Response(
                    httpx.codes.OK,
                    content=dummy_template_source.encode(),
                ),
            )
        get_workflow_route = respx_mock.get(f"{dummy_domain}{application_data.workflow_files[0].path}")
        get_workflow_route.mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_module_source.encode(),
            ),
        )

        desired_list_of_files = [
            tmp_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME,
            tmp_path / "templates" / "test-job-script.py.j2",
            tmp_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME,
        ]

        assert save_application_files(dummy_context, application_data, tmp_path) == desired_list_of_files
        assert set(tmp_path.rglob("*")) == {tmp_path / "templates", *desired_list_of_files}

        actual_application_config = load_application_config_from_source(
            (tmp_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME).read_text()
        )
        assert actual_application_config == application_config
        assert (tmp_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME).read_text() == dummy_module_source
        assert (tmp_path / "templates" / "test-job-script.py.j2").read_text() == dummy_template_source


class TestApplicationRuntime:
    @pytest.fixture(scope="function")
    def application_runtime(self, dummy_module_source, dummy_application_data):
        return ApplicationRuntime(
            ApplicationResponse.model_validate(dummy_application_data[0]),
            dummy_module_source,
            sdk=MagicMock(),
        )

    def test_flatten_param_dict__success(self, application_runtime):
        actual_result = application_runtime.as_flatten_param_dict()
        expected_result = {
            "template_files": None,
            "supporting_files_output_name": None,
            "supporting_files": None,
            "default_template": "test-job-script.py.j2",
            "foo": "bar",
        }

        assert actual_result == expected_result

    def test_get_app_config__success(self, application_runtime):
        expected_config = JobbergateApplicationConfig(
            application_config={"foo": "bar"},
            jobbergate_config={"default_template": "test-job-script.py.j2"},
        )

        assert application_runtime.app_config == expected_config

    def test_get_app_module__success(self, application_runtime):
        app_module = application_runtime.app_module
        assert isinstance(app_module, JobbergateApplicationBase)
        assert getattr(app_module, "mainflow", None) is not None
        assert app_module.sdk == application_runtime.sdk
        assert app_module.jobbergate_config == application_runtime.app_config.jobbergate_config.model_dump()
        assert app_module.application_config == application_runtime.app_config.application_config

    def test_execute_application__basic(
        self,
        application_runtime,
        dummy_render_class,
        mocker,
    ):
        dummy_render_class.prepared_input = dict(
            foo="FOO",
            bar="BAR",
            baz="BAZ",
        )
        mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)

        application_runtime.execute_application()
        assert application_runtime.answers == dict(
            foo="FOO",
            bar="BAR",
            baz="BAZ",
        )

    def test_execute_application__with_supplied_params(
        self,
        application_runtime,
        dummy_render_class,
        mocker,
    ):
        dummy_render_class.prepared_input = dict(
            foo="FOO",
            bar="BAR",
            baz="BAZ",
        )
        mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)

        application_runtime.supplied_params = dict(foo="oof")
        application_runtime.execute_application()
        assert application_runtime.answers == dict(
            foo="oof",
            bar="BAR",
            baz="BAZ",
        )

    def test_execute_application__with_fast_mode(
        self,
        application_runtime,
        dummy_render_class,
        mocker,
    ):
        dummy_render_class.prepared_input = dict(
            foo="FOO",
            bar="BAR",
            baz="BAZ",
        )
        mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)

        application_runtime.fast_mode = True

        application_runtime.execute_application()
        assert application_runtime.answers == dict(
            foo="FOO",
            bar="BAR",
            baz="zab",  # Only 'baz' has a default value, so it should be used
        )

    def test_set_name_dynamically(self, application_runtime):
        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                self.jobbergate_config["job_script_name"] = "very-unique-name"

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        application_runtime.execute_application()

        assert application_runtime.as_flatten_param_dict()["job_script_name"] == "very-unique-name"

    def test_set_name_dynamically__legacy(self, application_runtime):
        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                data["job_script_name"] = "very-unique-name"

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        application_runtime.execute_application()

        assert application_runtime.as_flatten_param_dict()["job_script_name"] == "very-unique-name"

    def test_choose_default_template(self, application_runtime):
        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                self.jobbergate_config["default_template"] = "very-unique-template"

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        application_runtime.execute_application()

        assert application_runtime.as_flatten_param_dict()["default_template"] == "very-unique-template"

    def test_choose_default_template__legacy(self, application_runtime):
        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                data["default_template"] = "very-unique-template"

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        application_runtime.execute_application()

        assert application_runtime.as_flatten_param_dict()["default_template"] == "very-unique-template"

    def test_gather_config_values__basic(self, application_runtime, dummy_render_class, mocker):
        variablename1 = "foo"
        question1 = Text(variablename1, message="gimme the foo!")

        variablename2 = "bar"
        question2 = Text(variablename2, message="gimme the bar!")

        variablename3 = "baz"
        question3 = Text(variablename3, message="gimme the baz!")

        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                data["nextworkflow"] = "subflow"
                return [question1, question2]

            def subflow(self, data):
                return [question3]

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        dummy_render_class.prepared_input = dict(
            foo="FOO",
            bar="BAR",
            baz="BAZ",
        )

        mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
        application_runtime._gather_answers()
        assert application_runtime.answers == dict(
            foo="FOO",
            bar="BAR",
            baz="BAZ",
        )

    def test_gather_config_values__returning_none(self, application_runtime, dummy_render_class, mocker):
        """
        Test that gather_param_values raises no error when a method returns None.

        Due to differences on the implementation details, jobbergate-legacy does not raise an error in this case
        and so legacy applications expect the same from next-gen Jobbergate.
        """
        variablename1 = "foo"
        question1 = Text(variablename1, message="gimme the foo!")

        variablename2 = "bar"
        question2 = Text(variablename2, message="gimme the bar!")

        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                data["nextworkflow"] = "subflow"
                return [question1, question2]

            def subflow(self, data):
                return None

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        dummy_render_class.prepared_input = dict(foo="FOO", bar="BAR", baz="BAZ")

        mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
        application_runtime._gather_answers()
        assert application_runtime.answers == dict(foo="FOO", bar="BAR")

    def test_gather_config_values__fast_mode(self, application_runtime, dummy_render_class, mocker):
        variablename1 = "foo"
        question1 = Text(variablename1, message="gimme the foo!", default="oof")

        variablename2 = "bar"
        question2 = Text(variablename2, message="gimme the bar!")

        variablename3 = "baz"
        question3 = Text(variablename3, message="gimme the baz!")

        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                data["nextworkflow"] = "subflow"
                return [question1, question2]

            def subflow(self, data):
                return [question3]

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        application_runtime.fast_mode = True
        dummy_render_class.prepared_input = dict(foo="FOO", bar="BAR", baz="BAZ")

        mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
        application_runtime._gather_answers()
        assert application_runtime.answers == dict(foo="oof", bar="BAR", baz="BAZ")

    def test_gather_config_values__skips_supplied_params(self, application_runtime, dummy_render_class, mocker):
        variablename1 = "foo"
        question1 = Text(variablename1, message="gimme the foo!", default="oof")

        variablename2 = "bar"
        question2 = Text(variablename2, message="gimme the bar!")

        variablename3 = "baz"
        question3 = Text(variablename3, message="gimme the baz!")

        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                data["nextworkflow"] = "subflow"
                return [question1, question2]

            def subflow(self, data):
                return [question3]

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        application_runtime.supplied_params = dict(bar="rab")
        dummy_render_class.prepared_input = dict(
            foo="FOO",
            bar="BAR",
            baz="BAZ",
        )

        mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
        application_runtime._gather_answers()
        assert application_runtime.answers == dict(foo="FOO", bar="rab", baz="BAZ")

    def test_gather_config_values__raises_Abort_if_method_not_implemented(
        self, application_runtime, dummy_render_class, mocker
    ):
        variablename1 = "foo"
        question1 = Text(variablename1, message="gimme the foo!")

        class DummyApplication(JobbergateApplicationBase):
            def mainflow(self, data):
                data["nextworkflow"] = "subflow"
                return [question1]

            def subflow(self, data):
                raise NotImplementedError("BOOM!")

        application_runtime.app_module = DummyApplication(application_runtime.app_config.model_dump())
        dummy_render_class.prepared_input = dict(
            foo="FOO",
        )

        mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
        with pytest.raises(Abort, match="not implemented"):
            application_runtime._gather_answers()



"""
Provide a ``typer`` app that can interact with Job Submission data in a cruddy manner.
"""

from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Optional, cast

import typer

from jobbergate_cli.constants import SortOrder
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.render import StyleMapper, render_single_result, terminal_message
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import ContextProtocol, JobSubmissionResponse
from jobbergate_cli.subapps.job_submissions.tools import fetch_job_submission_data, job_submissions_factory
from jobbergate_cli.subapps.pagination import handle_pagination
from jobbergate_cli.subapps.tools import resolve_selection


# move hidden field logic to the API
HIDDEN_FIELDS = [
    "cloned_from_id",
    "created_at",
    "execution_directory",
    "is_archived",
    "job_script",
    "report_message",
    "sbatch_arguments",
    "slurm_job_info",
    "updated_at",
]


style_mapper = StyleMapper(job_submission_id="green", name="cyan", slurm_job_id="dark_orange")


app = typer.Typer(
    rich_markup_mode="markdown",
    no_args_is_help=True,
    help=dedent(
        """Submit and track job submissions on Slurm clusters.

        Job submissions monitor the execution status and metadata of job scripts dispatched
        to Slurm clusters. They link job scripts to actual Slurm jobs, tracking:
        - Job status and state transitions (queued, running, completed, failed)
        - Slurm job IDs and job details

        Submissions can be made remotely or on-site (directly on the cluster).
        You can monitor progress, cancel running jobs, or clone submissions for re-execution.

        Workflow: applications → job-scripts → job-submissions

        Quick reference:
        - Create submission from application on submit mode: `jobbergate job-scripts create <application id or identifier> --submit`
        - Create from existing job script: `jobbergate job-submissions create <job_script_id>`
        - For full guide: `jobbergate --help`
        """
    ),
)


@app.command()
def create(
    ctx: typer.Context,
    job_script_id: Optional[int] = typer.Argument(
        None, help="The id of the job_script from which to create the job submission"
    ),
    name: str = typer.Option(
        ...,
        "--name",
        "-n",
        help="The name of the job submission to create",
    ),
    description: Optional[str] = typer.Option(
        None,
        help="A helpful description of the job submission",
    ),
    job_script_id_option: Optional[int] = typer.Option(
        None,
        "--job-script-id",
        "-i",
        help="Alternative way to specify the job script id",
    ),
    cluster_name: str = typer.Option(
        None,
        help="The name of the cluster where the job should be submitted (i.g. 'nash-staging')",
    ),
    execution_directory: Optional[Path] = typer.Option(
        None,
        help=dedent(
            """
            The path on the cluster where the job script should be executed.
            If provided as a relative path, it will be converted as an absolute path from your current
            working directory. If you use "~" to denote your home directory, the path will be expanded to an
            absolute path for your home directory on *this* machine.
            """
        ).strip(),
    ),
    sbatch_arguments: Optional[list[str]] = typer.Option(
        None,
        "--sbatch-arguments",
        "-s",
        help=dedent(
            """
            Additional arguments to pass as sbatch directives. These should be provided as a list of strings.
            See more details at: https://slurm.schedmd.com/sbatch.html
            """
        ).strip(),
    ),
    download: bool = typer.Option(
        False,
        help="Download the job script files to the current working directory",
    ),
):
    """
    Create a new job submission.
    """
    jg_ctx: ContextProtocol = ctx.obj
    job_script_id = resolve_selection(job_script_id, job_script_id_option)

    try:
        submissions_handler = job_submissions_factory(
            jg_ctx,
            job_script_id,
            name,
            description=description,
            execution_directory=execution_directory,
            cluster_name=cluster_name,
            sbatch_arguments=sbatch_arguments,
            download=download,
        )
        result = submissions_handler.run()
    except Exception as err:
        raise Abort(
            "Failed to create the job submission",
            subject="Job submission failed",
            support=True,
            log_message=f"There was an issue submitting the job from job_script_id={job_script_id}",
            original_error=err,
        )

    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Created Job Submission",
    )


@app.command("list")
def list_all(
    ctx: typer.Context,
    show_all: bool = typer.Option(
        False,
        "--all",
        help="Show all job submissions, even the ones owned by others",
    ),
    search: Optional[str] = typer.Option(None, help="Apply a search term to results"),
    sort_order: SortOrder = typer.Option(SortOrder.DESCENDING, help="Specify sort order"),
    sort_field: Optional[str] = typer.Option("id", help="The field by which results should be sorted"),
    from_job_script_id: Optional[int] = typer.Option(
        None,
        help="Filter job-submissions by the job-script-id they were created from.",
    ),
    include_archived: bool = typer.Option(False, "--include-archived", help="Include archived entries in the results"),
):
    """
    Show available job submissions.
    """
    jg_ctx: ContextProtocol = ctx.obj

    params: Dict[str, Any] = dict(user_only=not show_all, include_archived=include_archived)
    if search is not None:
        params["search"] = search
    if sort_order is not SortOrder.UNSORTED:
        params["sort_ascending"] = sort_order is SortOrder.ASCENDING
    if sort_field is not None:
        params["sort_field"] = sort_field
    if from_job_script_id is not None:
        params["from_job_script_id"] = from_job_script_id

    value_mappers = None
    organization_id = jg_ctx.authentication_handler.get_identity_data().organization_id
    if organization_id is not None:
        value_mappers = dict(cluster_name=lambda cn: cn.removesuffix(f"-{organization_id}"))

    handle_pagination(
        jg_ctx=jg_ctx,
        url_path="/jobbergate/job-submissions",
        abort_message="Couldn't retrieve job submissions list from API",
        params=params,
        title="Job Submission List",
        style_mapper=style_mapper,
        hidden_fields=HIDDEN_FIELDS,
        nested_response_model_cls=JobSubmissionResponse,
        value_mappers=value_mappers,
    )


@app.command()
def get_one(
    ctx: typer.Context,
    id: Optional[int] = typer.Argument(None, help="The specific id of the job submission to be selected."),
    id_option: Optional[int] = typer.Option(None, "--id", "-i", help="Alternative way to specify id"),
):
    """
    Show the detailed view of a single job submission by id
    """
    jg_ctx: ContextProtocol = ctx.obj
    id = resolve_selection(id, id_option)

    value_mappers = None
    organization_id = jg_ctx.authentication_handler.get_identity_data().organization_id
    if organization_id is not None:
        value_mappers = dict(cluster_name=lambda cn: cn.removesuffix(f"-{organization_id}"))

    result = fetch_job_submission_data(jg_ctx, id)
    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Job Submission",
        value_mappers=value_mappers,
    )


# NOTE: job submissions update is not added because it was an effective noop on the former implementation


@app.command()
def delete(
    ctx: typer.Context,
    id: Optional[int] = typer.Argument(
        None,
        help="The id of the job submission to delete",
    ),
    id_option: Optional[int] = typer.Option(
        ...,
        "--id",
        "-i",
        help="Alternative way to specify id",
    ),
):
    """
    Delete an existing job submission.
    """
    jg_ctx: ContextProtocol = ctx.obj
    id = resolve_selection(id, id_option)

    make_request(
        jg_ctx.client,
        f"/jobbergate/job-submissions/{id}",
        "DELETE",
        expected_status=204,
        abort_message="Request to delete job submission was not accepted by the API",
        support=True,
    )
    terminal_message(
        "The job submission was successfully deleted.",
        subject="Job submission delete succeeded",
    )


@app.command()
def clone(
    ctx: typer.Context,
    id: Optional[int] = typer.Argument(None, help="The specific id of the job submission to be updated."),
    id_option: Optional[int] = typer.Option(
        None,
        "--id",
        "-i",
        help="Alternative way to specify id.",
    ),
):
    """
    Clone an existing job submission under the CREATED status, so it is re-submitted to the cluster.
    """
    jg_ctx: ContextProtocol = ctx.obj
    id = resolve_selection(id, id_option)

    job_submission_result = cast(
        JobSubmissionResponse,
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-submissions/clone/{id}",
            "POST",
            expected_status=201,
            abort_message="Couldn't clone job submission",
            support=True,
            response_model_cls=JobSubmissionResponse,
        ),
    )

    render_single_result(
        jg_ctx,
        job_submission_result,
        hidden_fields=HIDDEN_FIELDS,
        title="Cloned Job Submission",
    )


@app.command()
def cancel(
    ctx: typer.Context,
    id: Optional[int] = typer.Argument(None, help="The specific id of the job submission to be cancelled."),
    id_option: Optional[int] = typer.Option(None, "--id", "-i", help="Alternative way to specify id"),
):
    """
    Cancel an existing job submission.
    """
    jg_ctx: ContextProtocol = ctx.obj
    id = resolve_selection(id, id_option)

    job_submission_result = cast(
        JobSubmissionResponse,
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-submissions/cancel/{id}",
            "PUT",
            expected_status=200,
            abort_message="Couldn't cancel job submission",
            support=True,
            response_model_cls=JobSubmissionResponse,
        ),
    )

    render_single_result(
        jg_ctx,
        job_submission_result,
        hidden_fields=HIDDEN_FIELDS,
        title="Cancelled Job Submission",
    )



import importlib
import json
import pathlib
import shlex
from unittest import mock

import httpx
import pytest

from jobbergate_cli.config import settings
from jobbergate_cli.schemas import ApplicationResponse, JobScriptFile, JobScriptResponse, JobSubmissionResponse
from jobbergate_cli.subapps.job_scripts.app import (
    HIDDEN_FIELDS,
    JOB_SUBMISSION_HIDDEN_FIELDS,
    clone,
    create,
    create_locally,
    create_stand_alone,
    delete,
    download_files,
    get_one,
    list_all,
    show_files,
    style_mapper,
    update,
)
from jobbergate_cli.text_tools import unwrap


def test_list_all__renders_paginated_results(
    make_test_app,
    dummy_context,
    cli_runner,
    mocker,
):
    test_app = make_test_app("list-all", list_all)
    mocked_pagination = mocker.patch("jobbergate_cli.subapps.job_scripts.app.handle_pagination")
    result = cli_runner.invoke(test_app, ["list-all"])
    assert result.exit_code == 0, f"list-all failed: {result.stdout}"
    mocked_pagination.assert_called_once_with(
        jg_ctx=dummy_context,
        url_path="/jobbergate/job-scripts",
        abort_message="Couldn't retrieve job scripts list from API",
        params={"user_only": True, "sort_ascending": False, "sort_field": "id", "include_archived": False},
        title="Job Scripts List",
        style_mapper=style_mapper,
        hidden_fields=HIDDEN_FIELDS,
        nested_response_model_cls=JobScriptResponse,
    )


@pytest.mark.parametrize("selector_template", ["{id}", "-i {id}", "--id={id}", "--id {id}"])
def test_get_one__success(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_job_script_data,
    dummy_domain,
    cli_runner,
    mocker,
    selector_template,
):
    job_script_data = dummy_job_script_data[0]
    id = job_script_data["id"]

    cli_selector = selector_template.format(id=id)

    respx_mock.get(f"{dummy_domain}/jobbergate/job-scripts/{id}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_job_script_data[0],
        ),
    )
    test_app = make_test_app("get-one", get_one)
    mocked_render = mocker.patch("jobbergate_cli.subapps.job_scripts.app.render_single_result")
    result = cli_runner.invoke(test_app, shlex.split(f"get-one {cli_selector}"))
    assert result.exit_code == 0, f"get-one failed: {result.stdout}"
    mocked_render.assert_called_once_with(
        dummy_context,
        JobScriptResponse.model_validate(dummy_job_script_data[0]),
        title="Job Script",
        hidden_fields=HIDDEN_FIELDS,
    )


def test_create_stand_alone__success(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_domain,
    dummy_render_class,
    cli_runner,
    tmp_path,
    attach_persona,
    mocker,
):
    dummy_job_script = tmp_path / "dummy.sh"
    dummy_job_script.write_text("echo hello world")

    dummy_support_1 = tmp_path / "dummy-support-1.txt"
    dummy_support_1.write_text("dummy 1")

    dummy_support_2 = tmp_path / "dummy-support-2.txt"
    dummy_support_2.write_text("dummy 2")

    create_route = respx_mock.post(f"{dummy_domain}/jobbergate/job-scripts")
    create_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=dict(
                id=1,
                created_at="2023-10-03 08:25:00",
                updated_at="2023-10-03 08:25:00",
                name="dummy-name",
                description=None,
                owner_email="dummy@dummy.com",
                application_id=None,
            ),
        )
    )

    upload_entrypoint_route = respx_mock.put(f"{dummy_domain}/jobbergate/job-scripts/1/upload/ENTRYPOINT")
    upload_entrypoint_route.mock(return_value=httpx.Response(httpx.codes.OK))

    upload_support_route = respx_mock.put(
        f"{dummy_domain}/jobbergate/job-scripts/1/upload/SUPPORT",
    )
    upload_support_route.mock(return_value=httpx.Response(httpx.codes.OK))

    attach_persona("dummy@dummy.com")

    test_app = make_test_app("create-stand-alone", create_stand_alone)

    mocked_render = mocker.patch("jobbergate_cli.subapps.job_scripts.app.render_single_result")
    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                create-stand-alone {dummy_job_script} --name=dummy-name
                       --supporting-file={dummy_support_1}
                       --supporting-file={dummy_support_2}
                """
            )
        ),
    )
    assert result.exit_code == 0, f"create failed: {result.stdout}"

    assert create_route.called
    content = json.loads(create_route.calls.last.request.content)
    assert content == {
        "name": "dummy-name",
        "description": None,
    }

    assert upload_entrypoint_route.call_count == 1
    assert b'filename="dummy.sh"' in upload_entrypoint_route.calls[0].request.content

    assert upload_support_route.call_count == 2
    assert b'filename="dummy-support-1.txt"' in upload_support_route.calls[0].request.content
    assert b'filename="dummy-support-2.txt"' in upload_support_route.calls[1].request.content

    mocked_render.assert_called_once_with(
        dummy_context,
        JobScriptResponse.model_validate(
            dict(
                id=1,
                created_at="2023-10-03 08:25:00",
                updated_at="2023-10-03 08:25:00",
                name="dummy-name",
                description=None,
                owner_email="dummy@dummy.com",
                application_id=None,
            ),
        ),
        title="Created Job Script",
        hidden_fields=HIDDEN_FIELDS,
    )


@pytest.mark.parametrize(
    "selector_template",
    ["{id}", "-i {id}", "--application-id={id}", "--application-id {id}"],
)
def test_create__non_fast_mode_and_job_submission(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_module_source,
    dummy_application_data,
    dummy_job_script_data,
    dummy_job_submission_data,
    dummy_domain,
    dummy_render_class,
    cli_runner,
    tmp_path,
    attach_persona,
    mocker,
    selector_template,
):
    application_response = ApplicationResponse(**dummy_application_data[0])
    id = application_response.application_id
    identifier = application_response.identifier

    url_selector = identifier if "identifier" in selector_template else id
    cli_selector = selector_template.format(id=id, identifier=identifier)

    job_script_data = dummy_job_script_data[0]

    job_submission_data = dummy_job_submission_data[0]

    render_route = respx_mock.post(f"{dummy_domain}/jobbergate/job-scripts/render-from-template/{url_selector}")
    render_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=job_script_data,
        ),
    )

    sbatch_params = " ".join(f"--sbatch-params={i}" for i in (1, 2, 3))

    param_file_path = tmp_path / "param_file.json"
    param_file_path.write_text(json.dumps(dict(foo="oof")))

    dummy_render_class.prepared_input = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )

    attach_persona("dummy@dummy.com")

    test_app = make_test_app("create", create)
    mocked_render = mocker.patch("jobbergate_cli.subapps.job_scripts.app.render_single_result")
    mocked_fetch_application_data = mocker.patch(
        "jobbergate_cli.subapps.job_scripts.tools.fetch_application_data",
        return_value=application_response,
    )
    assert len(application_response.workflow_files) >= 1
    get_workflow_route = respx_mock.get(f"{dummy_domain}{application_response.workflow_files[0].path}")
    get_workflow_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            content=dummy_module_source.encode(),
        ),
    )

    submissions_handler = mock.MagicMock()
    submissions_handler.run.return_value = JobSubmissionResponse.model_validate(job_submission_data)
    mocked_factory = mocker.patch(
        "jobbergate_cli.subapps.job_scripts.app.job_submissions_factory", return_value=submissions_handler
    )

    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                create {cli_selector}
                       --name dummy-name
                       --param-file={param_file_path}
                       {sbatch_params}
                """
            )
        ),
        # To confirm that the job should be submitted to the default cluster, in the current dir and not downloaded
        input=f"y\nn\n{settings.DEFAULT_CLUSTER_NAME}\n.\n",
    )
    assert result.exit_code == 0, f"create failed: {result.stdout}"
    mocked_fetch_application_data.assert_called_once_with(
        dummy_context,
        application_response.application_id,
    )

    assert render_route.call_count == 1
    content = json.loads(render_route.calls.last.request.content)
    assert content == {
        "create_request": {"name": "dummy-name", "description": None},
        "render_request": {
            "template_output_name_mapping": {"test-job-script.py.j2": "test-job-script.py"},
            "sbatch_params": ["1", "2", "3"],
            "param_dict": {
                "data": {
                    "foo": "oof",
                    "bar": "BAR",
                    "baz": "BAZ",
                    "template_files": None,
                    "default_template": "test-job-script.py.j2",
                    "supporting_files_output_name": None,
                    "supporting_files": None,
                }
            },
        },
    }

    mocked_factory.assert_called_once_with(
        jg_ctx=dummy_context,
        job_script_id=job_script_data["id"],
        name=job_script_data["name"],
        description=job_script_data["description"],
        cluster_name=None,
        execution_directory=None,
        sbatch_arguments=None,
    )

    mocked_render.assert_has_calls(
        [
            mocker.call(
                dummy_context,
                JobScriptResponse(**job_script_data),
                title="Created Job Script",
                hidden_fields=HIDDEN_FIELDS,
            ),
            mocker.call(
                dummy_context,
                JobSubmissionResponse(**job_submission_data),
                title="Created Job Submission (Fast Mode)",
                hidden_fields=JOB_SUBMISSION_HIDDEN_FIELDS,
            ),
        ]
    )


def test_create__with_fast_mode_and_no_job_submission(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_module_source,
    dummy_application_data,
    dummy_job_script_data,
    dummy_domain,
    cli_runner,
    tmp_path,
    attach_persona,
    mocker,
):
    application_response = ApplicationResponse(**dummy_application_data[0])

    job_script_data = dummy_job_script_data[0]

    render_route = respx_mock.post(
        f"{dummy_domain}/jobbergate/job-scripts/render-from-template/{application_response.application_id}"
    )
    render_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=job_script_data,
        ),
    )

    sbatch_params = " ".join(f"--sbatch-params={i}" for i in (1, 2, 3))

    param_file_path = tmp_path / "param_file.json"
    param_file_path.write_text(
        json.dumps(
            dict(
                foo="oof",
                bar="rab",
                baz="zab",
            )
        )
    )

    attach_persona("dummy@dummy.com")

    test_app = make_test_app("create", create)
    mocked_render = mocker.patch("jobbergate_cli.subapps.job_scripts.app.render_single_result")
    mocked_fetch_application_data = mocker.patch(
        "jobbergate_cli.subapps.job_scripts.tools.fetch_application_data",
        return_value=application_response,
    )
    assert len(application_response.workflow_files) >= 1
    get_workflow_route = respx_mock.get(f"{dummy_domain}{application_response.workflow_files[0].path}")
    get_workflow_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            content=dummy_module_source.encode(),
        ),
    )
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                create --name=dummy-name
                       --application-id={application_response.application_id}
                       --param-file={param_file_path}
                       --fast
                       --no-submit
                       --no-download
                       {sbatch_params}
                """
            )
        ),
    )
    assert result.exit_code == 0, f"create failed: {result.stdout}"
    mocked_fetch_application_data.assert_called_once_with(dummy_context, application_response.application_id)
    assert render_route.call_count == 1
    content = json.loads(render_route.calls.last.request.content)
    assert content == {
        "create_request": {"name": "dummy-name", "description": None},
        "render_request": {
            "template_output_name_mapping": {"test-job-script.py.j2": "test-job-script.py"},
            "sbatch_params": ["1", "2", "3"],
            "param_dict": {
                "data": {
                    "foo": "oof",
                    "bar": "rab",
                    "baz": "zab",
                    "template_files": None,
                    "default_template": "test-job-script.py.j2",
                    "supporting_files_output_name": None,
                    "supporting_files": None,
                }
            },
        },
    }

    mocked_render.assert_called_once_with(
        dummy_context,
        JobScriptResponse(**job_script_data),
        title="Created Job Script",
        hidden_fields=HIDDEN_FIELDS,
    )


def test_create__submit_is_none_and_cluster_name_is_defined(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_application_data,
    dummy_job_script_data,
    dummy_domain,
    cli_runner,
    tmp_path,
    attach_persona,
    mocker,
):
    application_response = ApplicationResponse(**dummy_application_data[0])

    job_script_data = dummy_job_script_data[0]

    render_route = respx_mock.post(
        f"{dummy_domain}/jobbergate/job-scripts/render-from-template/{application_response.application_id}"
    )
    render_route.mock(
        return_value=httpx.Response(httpx.codes.CREATED, json=job_script_data),
    )

    sbatch_params = " ".join(f"--sbatch-params={i}" for i in (1, 2, 3))

    param_file_path = tmp_path / "param_file.json"
    param_file_path.write_text(
        json.dumps(
            dict(
                foo="oof",
                bar="rab",
                baz="zab",
            )
        )
    )

    attach_persona("dummy@dummy.com")

    test_app = make_test_app("create", create)
    mocked_render = mocker.patch("jobbergate_cli.subapps.job_scripts.app.render_single_result")
    mocked_fetch_application_data = mocker.patch(
        "jobbergate_cli.subapps.job_scripts.app.render_job_script",
        return_value=JobScriptResponse(**job_script_data),
    )

    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                create --name=dummy-name
                       --application-id={application_response.application_id}
                       --param-file={param_file_path}
                       --cluster-name=dummy-cluster
                       {sbatch_params}
                """
            )
        ),
    )
    assert mocked_fetch_application_data.call_count == 1
    mocked_render.assert_called_once_with(
        dummy_context,
        JobScriptResponse(**job_script_data),
        title="Created Job Script",
        hidden_fields=HIDDEN_FIELDS,
    )

    assert result.exit_code == 1, f"create failed: {result.stdout}"
    assert "Incorrect parameters" in result.stdout


def test_create_job_script_locally__success(
    dummy_render_class, dummy_application_dir, make_test_app, cli_runner, mocker
):
    test_app = make_test_app("create-locally", create_locally)

    dummy_render_class.prepared_input = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )
    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )

    mocked_terminal_message = mocker.patch("jobbergate_cli.subapps.job_scripts.app.terminal_message")

    result = cli_runner.invoke(
        test_app,
        shlex.split(f"create-locally {dummy_application_dir} --output-path={dummy_application_dir}"),
    )

    assert result.exit_code == 0, f"create-locally failed: {result.stdout}"

    mocked_terminal_message.assert_called_once_with(
        "The job script was successfully rendered locally.",
        subject="Job script render succeeded",
    )


@pytest.mark.parametrize("selector_template", ["{id}", "-i {id}", "--id={id}", "--id {id}"])
def test_update__makes_request_and_renders_results(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_job_script_data,
    dummy_domain,
    cli_runner,
    mocker,
    selector_template,
):
    job_script_data = dummy_job_script_data[0]
    job_script_id = job_script_data["id"]

    cli_selector = selector_template.format(id=job_script_id)

    new_job_script_data = {
        **job_script_data,
        "name": "new-test-name",
        "description": "new-test-description",
    }
    respx_mock.put(f"{dummy_domain}/jobbergate/job-scripts/{job_script_id}").mock(
        return_value=httpx.Response(httpx.codes.OK, json=new_job_script_data),
    )
    test_app = make_test_app("update", update)
    mocked_render = mocker.patch("jobbergate_cli.subapps.job_scripts.app.render_single_result")
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                update {cli_selector}
                       --name='new-test-name'
                       --description='new-test-description'
                """
            )
        ),
    )
    assert result.exit_code == 0, f"update failed: {result.stdout}"
    mocked_render.assert_called_once_with(
        dummy_context,
        JobScriptResponse(**new_job_script_data),
        title="Updated Job Script",
        hidden_fields=HIDDEN_FIELDS,
    )


@pytest.mark.parametrize("selector_template", ["{id}", "-i {id}", "--id={id}", "--id {id}"])
def test_delete__makes_request_and_sends_terminal_message(
    respx_mock,
    make_test_app,
    dummy_domain,
    cli_runner,
    selector_template,
):
    job_script_id = 13

    cli_selector = selector_template.format(id=job_script_id)

    delete_route = respx_mock.delete(f"{dummy_domain}/jobbergate/job-scripts/{job_script_id}").mock(
        return_value=httpx.Response(httpx.codes.NO_CONTENT),
    )
    test_app = make_test_app("delete", delete)
    result = cli_runner.invoke(test_app, shlex.split(f"delete {cli_selector}"))
    assert result.exit_code == 0, f"delete failed: {result.stdout}"
    assert delete_route.called
    assert "JOB SCRIPT DELETE SUCCEEDED"


@pytest.mark.parametrize("selector_template", ["{id}", "-i {id}", "--id={id}", "--id {id}"])
def test_show_files__success(
    respx_mock,
    make_test_app,
    dummy_job_script_data,
    dummy_domain,
    dummy_template_source,
    cli_runner,
    mocker,
    selector_template,
):
    """
    Verify that the ``show-files`` subcommand works as expected.
    """
    job_script_data = dummy_job_script_data[0]
    id = job_script_data["id"]

    cli_selector = selector_template.format(id=id)

    respx_mock.get(f"{dummy_domain}/jobbergate/job-scripts/{id}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=job_script_data,
        ),
    )

    get_file_routes = [
        respx_mock.get(f"{dummy_domain}{JobScriptFile.model_validate(f).path}") for f in job_script_data["files"]
    ]
    for route in get_file_routes:
        route.mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                content=dummy_template_source.encode(),
            ),
        )

    test_app = make_test_app("show-files", show_files)
    mocked_terminal_message = mocker.patch("jobbergate_cli.subapps.job_scripts.app.terminal_message")

    result = cli_runner.invoke(test_app, shlex.split(f"show-files {cli_selector}"))
    assert result.exit_code == 0, f"get-one failed: {result.stdout}"
    mocked_terminal_message.assert_called_once_with(
        dummy_template_source,
        subject="application.sh",
        footer="This is the main job script file",
    )


class TestDownloadJobScriptFiles:
    """
    Test the ``download`` subcommand.
    """

    @pytest.fixture()
    def test_app(self, make_test_app):
        """
        Fixture to create a test app.
        """
        return make_test_app("download", download_files)

    @pytest.mark.parametrize("selector_template", ["{id}", "-i {id}", "--id={id}", "--id {id}"])
    def test_download__success(
        self,
        respx_mock,
        test_app,
        dummy_job_script_data,
        dummy_domain,
        dummy_context,
        cli_runner,
        mocker,
        tmp_path,
        selector_template,
    ):
        """
        Test that the ``download`` subcommand works as expected.
        """
        job_script_data = dummy_job_script_data[0]
        id = job_script_data["id"]

        cli_selector = selector_template.format(id=id)

        respx_mock.get(f"{dummy_domain}/jobbergate/job-scripts/{id}").mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                json=job_script_data,
            ),
        )
        mocked_render = mocker.patch("jobbergate_cli.subapps.job_scripts.app.terminal_message")

        with mock.patch.object(pathlib.Path, "cwd", return_value=tmp_path):
            with mock.patch(
                "jobbergate_cli.subapps.job_scripts.app.download_job_script_files",
                return_value=list(f["filename"] for f in job_script_data["files"]),
            ) as mocked_save_job_script_files:
                result = cli_runner.invoke(test_app, shlex.split(f"download {cli_selector}"))

                mocked_save_job_script_files.assert_called_once_with(1, dummy_context, tmp_path)

        assert result.exit_code == 0, f"download failed: {result.stdout}"
        mocked_render.assert_called_once_with(
            "A total of 1 job script files were successfully downloaded.",
            subject="Job script download succeeded",
        )


class TestCloneJobScript:
    @pytest.mark.parametrize("selector_template", ["{id}", "-i {id}", "--id={id}", "--id {id}"])
    def test_clone__success(
        self,
        respx_mock,
        make_test_app,
        dummy_job_script_data,
        dummy_domain,
        dummy_context,
        cli_runner,
        mocker,
        selector_template,
    ):
        """
        Test that the clone application subcommand works as expected.
        """

        job_script_data = dummy_job_script_data[0]
        id = job_script_data["id"]

        cli_selector = selector_template.format(id=id)

        clone_route = respx_mock.post(f"{dummy_domain}/jobbergate/job-scripts/clone/{id}").mock(
            return_value=httpx.Response(
                httpx.codes.CREATED,
                json=job_script_data,
            ),
        )
        mocked_render = mocker.patch("jobbergate_cli.subapps.job_scripts.app.render_single_result")

        test_app = make_test_app("clone", clone)
        result = cli_runner.invoke(
            test_app,
            shlex.split(
                "clone {} --name={} --description={}".format(
                    cli_selector,
                    shlex.quote(job_script_data["name"]),
                    shlex.quote(job_script_data["description"]),
                ),
            ),
        )

        assert clone_route.called

        assert result.exit_code == 0, f"clone failed: {result.stdout}"
        mocked_render.assert_called_once_with(
            dummy_context,
            JobScriptResponse(**job_script_data),
            title="Cloned Job Script",
            hidden_fields=HIDDEN_FIELDS,
        )



import pytest
import typer

from jobbergate_cli.exceptions import Abort, handle_abort


@pytest.fixture
def dummy_exception():
    return RuntimeError("Detonate!")


@pytest.fixture
def dummy_handled_function(dummy_exception):
    @handle_abort
    def _wrapped(
        message: str = "Explosive Message",
        subject: str = "BOOM!",
        support: bool = True,
        log_message: str = "volatile log message",
        original_error: Exception = dummy_exception,
        add_sentry_context: bool = True,
    ):
        sentry_kwarg = dict(sentry_context=dict(blast=True)) if add_sentry_context else dict()
        raise Abort(
            message,
            subject=subject,
            support=support,
            log_message=log_message,
            original_error=original_error,
            **sentry_kwarg,
        )

    return _wrapped


def test_handle_abort__with_all_options(capsys, caplog, mocker, dummy_handled_function, dummy_exception):
    mocked_sentry_capture = mocker.patch("jobbergate_cli.exceptions.sentry_sdk.capture_exception")
    with pytest.raises(typer.Exit):
        dummy_handled_function()

    captured = capsys.readouterr()
    assert "Explosive Message" in captured.out
    assert "BOOM!" in captured.out
    assert "If the problem persists" in captured.out

    assert "volatile log message" in caplog.text
    assert f"Original exception: {dummy_exception}" in caplog.text
    assert mocked_sentry_capture.call_count == 0


def test_handle_abort__without_SENTRY_DNS_does_not_push_to_sentry(mocker, tweak_settings, dummy_handled_function):
    mocked_sentry_capture = mocker.patch("jobbergate_cli.exceptions.sentry_sdk.capture_exception")
    with tweak_settings(SENTRY_DSN=None):
        with pytest.raises(typer.Exit):
            dummy_handled_function()

    assert not mocked_sentry_capture.called


def test_handle_abort__does_not_log_if_log_message_and_original_error_are_None(
    caplog, tweak_settings, dummy_handled_function
):
    with tweak_settings(SENTRY_DSN=None):
        with pytest.raises(typer.Exit):
            dummy_handled_function(log_message=None, original_error=None)

    assert caplog.text == ""


def test_handle_abort__does_not_include_support_message_if_support_is_False(capsys, dummy_handled_function):
    with pytest.raises(typer.Exit):
        dummy_handled_function(support=False)

    captured = capsys.readouterr()
    assert "If the problem persists" not in captured.out


def test_handle_abort__does_not_include_subject_message_if_subject_is_None(mocker, dummy_handled_function):
    mocked_panel = mocker.patch("jobbergate_cli.exceptions.Panel")
    with pytest.raises(typer.Exit):
        dummy_handled_function(message="Bang!", subject=None)

    assert mocked_panel.call_count == 1



import json
from datetime import datetime, timezone

import httpx
import plummet
import pytest

from jobbergate_cli.config import settings
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.schemas import ClusterCacheData
from jobbergate_cli.subapps.clusters.tools import (
    get_client_ids,
    load_clusters_from_cache,
    pull_client_ids_from_api,
    save_clusters_to_cache,
)


@pytest.fixture
def dummy_domain():
    return "https://dummy.com"


def test_pull_client_ids_from_api__success(respx_mock, dummy_domain, dummy_context):
    clusters_route = respx_mock.post(f"{dummy_domain}/cluster/graphql/query")
    clusters_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(
                data=dict(
                    cluster=[
                        dict(clientId="cluster1"),
                        dict(clientId="cluster2"),
                        dict(clientId="cluster3"),
                    ],
                ),
            ),
        ),
    )

    assert pull_client_ids_from_api(dummy_context) == ["cluster1", "cluster2", "cluster3"]


def test_pull_client_ids_from_api__raises_abort_on_non_200(respx_mock, dummy_domain, dummy_context):
    clusters_route = respx_mock.post(f"{dummy_domain}/cluster/graphql/query")
    clusters_route.mock(
        return_value=httpx.Response(
            httpx.codes.BAD_REQUEST,
        ),
    )

    with pytest.raises(Abort, match="There was a problem retrieving registered clusters"):
        pull_client_ids_from_api(dummy_context)


def test_pull_client_ids_from_api__raises_abort_on_malformed_response(respx_mock, dummy_domain, dummy_context):
    clusters_route = respx_mock.post(f"{dummy_domain}/cluster/graphql/query")
    clusters_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(bad="data"),
        ),
    )

    with pytest.raises(Abort, match="Couldn't unpack cluster names"):
        pull_client_ids_from_api(dummy_context)


def test_save_clusters_to_cache(tmp_path, tweak_settings):
    with tweak_settings(JOBBERGATE_CACHE_DIR=tmp_path):
        with plummet.frozen_time("2022-05-13 16:56:00"):
            save_clusters_to_cache(["cluster1", "cluster2", "cluster3"])

        cache_data = ClusterCacheData(**json.loads(settings.JOBBERGATE_CLUSTER_LIST_PATH.read_text()))

    assert cache_data.client_ids == ["cluster1", "cluster2", "cluster3"]
    assert plummet.moments_match(cache_data.updated_at, "2022-05-13 16:56:00")


def test_load_clusters_from_cache__success(tmp_path, tweak_settings):
    with tweak_settings(JOBBERGATE_CACHE_DIR=tmp_path, JOBBERGATE_CLUSTER_CACHE_LIFETIME=5):
        with plummet.frozen_time("2022-05-13 16:56:00"):
            cache_data = ClusterCacheData(
                updated_at=datetime.now(timezone.utc),
                client_ids=["cluster1", "cluster2", "cluster3"],
            )
            settings.JOBBERGATE_CLUSTER_LIST_PATH.write_text(cache_data.model_dump_json())

            assert load_clusters_from_cache() == ["cluster1", "cluster2", "cluster3"]


def test_load_clusters_from_cache__returns_None_if_cache_is_expired(tmp_path, tweak_settings):
    with tweak_settings(JOBBERGATE_CACHE_DIR=tmp_path, JOBBERGATE_CLUSTER_CACHE_LIFETIME=5):
        with plummet.frozen_time("2022-05-13 16:56:00"):
            cache_data = ClusterCacheData(
                updated_at=datetime.now(timezone.utc),
                client_ids=["cluster1", "cluster2", "cluster3"],
            )
            settings.JOBBERGATE_CLUSTER_LIST_PATH.write_text(cache_data.model_dump_json())

        with plummet.frozen_time("2022-05-13 16:56:06"):
            assert load_clusters_from_cache() is None


def test_load_clusters_from_cache__returns_None_if_cache_is_invalid(tmp_path, tweak_settings):
    with tweak_settings(JOBBERGATE_CACHE_DIR=tmp_path, JOBBERGATE_CLUSTER_CACHE_LIFETIME=5):
        settings.JOBBERGATE_CLUSTER_LIST_PATH.write_text("BAD DATA")
        assert load_clusters_from_cache() is None


def test_get_client_ids__pulls_from_api_if_no_cache_available(
    tmp_path, mocker, respx_mock, dummy_domain, dummy_context, tweak_settings
):
    """

    Also assert that a new cache file was created.
    """
    mocker.patch("jobbergate_cli.subapps.clusters.tools.load_clusters_from_cache", return_value=None)
    clusters_route = respx_mock.post(f"{dummy_domain}/cluster/graphql/query")
    clusters_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(
                data=dict(
                    cluster=[
                        dict(clientId="cluster1"),
                        dict(clientId="cluster2"),
                        dict(clientId="cluster3"),
                    ],
                ),
            ),
        ),
    )

    with tweak_settings(JOBBERGATE_CACHE_DIR=tmp_path):
        with plummet.frozen_time("2022-05-16 15:38:00"):
            assert get_client_ids(dummy_context) == ["cluster1", "cluster2", "cluster3"]
        assert clusters_route.called
        cached_data = json.loads(settings.JOBBERGATE_CLUSTER_LIST_PATH.read_text())
        assert cached_data["client_ids"] == ["cluster1", "cluster2", "cluster3"]
        assert plummet.moments_match(cached_data["updated_at"], "2022-05-16 15:38:00")


def test_get_client_ids__loads_from_cache_when_available(mocker, respx_mock, dummy_domain, dummy_context):
    mocker.patch("jobbergate_cli.subapps.clusters.tools.load_clusters_from_cache", return_value=None)
    clusters_route = respx_mock.post(f"{dummy_domain}/cluster/graphql/query")
    clusters_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(
                data=dict(
                    cluster=[
                        dict(clientId="cluster1"),
                        dict(clientId="cluster2"),
                        dict(clientId="cluster3"),
                    ],
                ),
            ),
        ),
    )

    assert get_client_ids(dummy_context) == ["cluster1", "cluster2", "cluster3"]
    assert clusters_route.called



import json
from typing import Dict, Optional

import httpx
import pydantic
import pytest

from jobbergate_cli.exceptions import Abort
from jobbergate_cli.requests import _deserialize_request_model, make_request
from jobbergate_cli.schemas import ListResponseEnvelope


DEFAULT_DOMAIN = "https://dummy-domain.com"


@pytest.fixture
def dummy_client():
    """
    Provide factory for a test client. Can supply custom base_url and headers.
    """

    def _helper(base_url: str = DEFAULT_DOMAIN, headers: Optional[Dict] = None) -> httpx.Client:
        """
        Create the dummy httpx client.
        """
        if headers is None:
            headers = dict()

        return httpx.Client(base_url=base_url, headers=headers)

    return _helper


class DummyResponseModel(pydantic.BaseModel):
    """
    Provide a dummy pydantic model for testing standard responses.
    """

    foo: int
    bar: str


class ErrorResponseModel(pydantic.BaseModel):
    """
    Provide a dummy pydantic model for testing error responses.
    """

    error: str


def test__deserialize_request_model__success():
    """
    Validate that the ``_deserialize_request_model`` method can successfully deserialize a pydantic model instance into
    the ``content`` part of the ``request_kwargs``. Also, validate that the ``content-type`` part of the request is set
    to ``application/json``.
    """
    request_kwargs = dict()
    _deserialize_request_model(
        DummyResponseModel(foo=1, bar="one"),
        request_kwargs,
        "Abort message does not matter here",
        "Whatever Subject",
    )
    assert json.loads(request_kwargs["content"]) == dict(foo=1, bar="one")
    assert request_kwargs["headers"] == {"Content-Type": "application/json"}


def test__deserialize_request_model__raises_Abort_if_request_kwargs_already_has_other_body_parts():
    """
    Validate that the ``_deserialize_request_model`` raises an Abort if the ``request_kwargs`` already has a "body" part
    (``data``, ``json``, or ``content``).
    """
    with pytest.raises(Abort, match="Request was incorrectly structured"):
        _deserialize_request_model(
            DummyResponseModel(foo=1, bar="one"),
            dict(data=dict(foo=11)),
            "Abort message does not matter here",
            "Whatever Subject",
        )

    with pytest.raises(Abort, match="Request was incorrectly structured"):
        _deserialize_request_model(
            DummyResponseModel(foo=1, bar="one"),
            dict(json=dict(foo=11)),
            "Abort message does not matter here",
            "Whatever Subject",
        )

    with pytest.raises(Abort, match="Request was incorrectly structured"):
        _deserialize_request_model(
            DummyResponseModel(foo=1, bar="one"),
            dict(content=json.dumps(dict(foo=11))),
            "Abort message does not matter here",
            "Whatever Subject",
        )


def test_make_request__success(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will make a request to the supplied path for the
    client domain, check the status code, extract the data, validate it, and return an instance of the
    supplied response model.
    """

    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(
                foo=1,
                bar="one",
            ),
        ),
    )
    dummy_response_instance = make_request(
        client,
        req_path,
        "GET",
        expected_status=200,
        response_model_cls=DummyResponseModel,
    )
    assert isinstance(dummy_response_instance, DummyResponseModel)
    assert dummy_response_instance.foo == 1
    assert dummy_response_instance.bar == "one"


def test_make_request__raises_Abort_if_client_request_raises_exception(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will raise an Abort if the call to ``client.send`` raises an
    exception.
    """
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"
    original_error = httpx.RequestError("BOOM!")

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(side_effect=original_error)

    with pytest.raises(Abort, match="There was a big problem: Communication with the API failed") as err_info:
        make_request(
            client, req_path, "GET", abort_message="There was a big problem", abort_subject="BIG PROBLEM", support=True
        )
    assert err_info.value.subject == "BIG PROBLEM - RequestError"
    assert err_info.value.support is True
    assert err_info.value.log_message == "There was an error on the request -- BOOM!"
    assert err_info.value.original_error == original_error


def test_make_request__raises_Abort_with_ownership_message_for_403_for_non_owners(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will raise an Abort if the ``expected_status`` arg is set and it
    does not match the status code of the response. Further verify that message is attached telling the user that
    they may not modify a resource they do not own.
    """
    client = dummy_client()
    req_path = "/fake-path"

    respx_mock.delete(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.FORBIDDEN,
            json=dict(detail="This jabroni does not own this whingding"),
        ),
    )

    with pytest.raises(
        Abort,
        match="There was a big problem -- This jabroni does not own this whingding -- Please verify your credentials",
    ) as err_info:
        make_request(
            client,
            req_path,
            "DELETE",
            expected_status=204,
            abort_message="There was a big problem",
            abort_subject="BIG PROBLEM",
            support=True,
        )
    assert err_info.value.subject == "BIG PROBLEM - Forbidden"
    assert err_info.value.support is False
    assert "Request was invalid due to a client-side error" in err_info.value.log_message
    assert "This jabroni does not own this whingding" in err_info.value.log_message
    assert err_info.value.original_error is None


def test_make_request__raises_Abort_when_expected_status_is_not_None_and_response_status_does_not_match_it(
    respx_mock, dummy_client
):
    """
    Validate that the ``make_request()`` function will raise an Abort if the ``expected_status`` arg is set and it
    does not match the status code of the response.
    """
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.BAD_REQUEST,
            text="It blowed up",
        ),
    )

    with pytest.raises(
        Abort, match="There was a big problem -- Please check the data on your request and try again"
    ) as err_info:
        make_request(
            client,
            req_path,
            "GET",
            expected_status=200,
            abort_message="There was a big problem",
            abort_subject="BIG PROBLEM",
            support=True,
        )
    assert err_info.value.subject == "BIG PROBLEM - Bad Request"
    assert err_info.value.support is True
    assert (
        err_info.value.log_message
        == "Request was invalid due to a client-side error (400 -- Bad Request): It blowed up"
    )
    assert err_info.value.original_error is None


def test_make_request__does_not_raise_Abort_when_expected_status_is_None_and_response_status_is_a_fail_code(
    respx_mock, dummy_client
):
    """
    Validate that the ``make_request()`` function will not raise an Abort if the ``expected_status`` arg is not set
    and the response is an error status code.
    """
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.BAD_REQUEST,
            json=dict(error="It blowed up"),
        ),
    )

    err = make_request(
        client,
        req_path,
        "GET",
        response_model_cls=ErrorResponseModel,
    )
    assert err.error == "It blowed up"


def test_make_request__returns_the_response_status_code_if_the_method_is_DELETE(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will return None if the ``method`` arg is DELETE and the request
    was successfull.
    """
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.delete(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(error="It blowed up"),
        ),
    )

    assert make_request(client, req_path, "DELETE") == httpx.codes.OK


def test_make_request__returns_the_response_status_code_if_expect_response_is_False(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will return None if the ``expect_response`` arg is False and the
    request was successfull.
    """
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.post(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(httpx.codes.BAD_REQUEST),
    )

    assert make_request(client, req_path, "POST", expect_response=False) == httpx.codes.BAD_REQUEST


def test_make_request__raises_an_Abort_if_the_response_cannot_be_deserialized_with_JSON(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will raise an Abort if the response is not JSON de-serializable.
    """
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            text="Not JSON, my dude",
        ),
    )

    with pytest.raises(Abort, match="There was a big problem: Response carried no data") as err_info:
        make_request(
            client, req_path, "GET", abort_message="There was a big problem", abort_subject="BIG PROBLEM", support=True
        )
    assert err_info.value.subject == "BIG PROBLEM"
    assert err_info.value.support is True
    assert err_info.value.log_message == "Failed unpacking json: Not JSON, my dude"
    assert isinstance(err_info.value.original_error, json.decoder.JSONDecodeError)


def test_make_request__returns_a_plain_dict_if_response_model_cls_is_None(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will return a plain dictionary containing the response data if the
    ``response_model_cls`` argument is not supplied.
    """
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(a=1, b=2, c=3),
        ),
    )

    assert make_request(client, req_path, "GET") == dict(a=1, b=2, c=3)


def test_make_request__raises_an_Abort_if_the_response_data_cannot_be_serialized_into_the_response_model_cls(
    respx_mock, dummy_client
):
    """
    Validate that the ``make_request()`` function will raise an Abort if the response data cannot be serialized as and
    validated with the ``response_model_cls``.
    """
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(a=1, b=2, c=3),
        ),
    )

    with pytest.raises(Abort, match="There was a big problem: Unexpected data in response") as err_info:
        make_request(
            client,
            req_path,
            "GET",
            abort_message="There was a big problem",
            abort_subject="BIG PROBLEM",
            support=True,
            response_model_cls=DummyResponseModel,
        )
    assert err_info.value.subject == "BIG PROBLEM"
    assert err_info.value.support is True
    assert err_info.value.log_message == f"Unexpected format in response data: {dict(a=1, b=2, c=3)}"
    assert isinstance(err_info.value.original_error, pydantic.ValidationError)


def test_make_request__uses_request_model_instance_for_request_body_if_passed(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will use a pydantic model instance to build the body of the request if
    the ``request_model`` argument is passed.
    """

    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    dummy_route = respx_mock.post(f"{DEFAULT_DOMAIN}{req_path}")
    dummy_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=dict(
                foo=1,
                bar="one",
            ),
        ),
    )
    dummy_response_instance = make_request(
        client,
        req_path,
        "POST",
        expected_status=201,
        response_model_cls=DummyResponseModel,
        request_model=DummyResponseModel(foo=1, bar="one"),
    )
    assert isinstance(dummy_response_instance, DummyResponseModel)
    assert dummy_response_instance.foo == 1
    assert dummy_response_instance.bar == "one"

    assert json.loads(dummy_route.calls.last.request.content.decode()) == dict(foo=1, bar="one")
    assert dummy_route.calls.last.request.headers["Content-Type"] == "application/json"


def test_make_request__can_use_unpack_response_into_ListResponseEnvelope(respx_mock, dummy_client):
    """
    Validate that the ``make_request()`` function will use a pydantic model instance to build the body of the request if
    the ``request_model`` argument is passed.
    """

    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    dummy_route = respx_mock.post(f"{DEFAULT_DOMAIN}{req_path}")
    dummy_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=dict(
                items=[
                    dict(foo=1, bar="one"),
                    dict(foo=2, bar="two"),
                    dict(foo=3, bar="three"),
                ],
                total=3,
                page=0,
                size=5,
                pages=1,
            ),
        ),
    )
    dummy_response_instance = make_request(
        client,
        req_path,
        "POST",
        expected_status=201,
        response_model_cls=ListResponseEnvelope[DummyResponseModel],
        request_model=DummyResponseModel(foo=1, bar="one"),
    )
    assert isinstance(dummy_response_instance, ListResponseEnvelope)
    assert dummy_response_instance.items == [
        DummyResponseModel(foo=1, bar="one"),
        DummyResponseModel(foo=2, bar="two"),
        DummyResponseModel(foo=3, bar="three"),
    ]
    assert dummy_response_instance.total == 3
    assert dummy_response_instance.page == 0
    assert dummy_response_instance.size == 5
    assert dummy_response_instance.pages == 1

    assert json.loads(dummy_route.calls.last.request.content.decode()) == dict(foo=1, bar="one")
    assert dummy_route.calls.last.request.headers["Content-Type"] == "application/json"



from pathlib import Path
import pytest
from click import UsageError
from jobbergate_cli.subapps.tools import resolve_application_selection, resolve_selection


class TestResolveApplicationSelection:
    def test_resolve_application_selection_digit(self):
        assert resolve_application_selection(id_or_identifier="123") == 123

    def test_resolve_application_selection_string(self):
        assert resolve_application_selection(id_or_identifier="app-identifier") == "app-identifier"

    def test_resolve_application_selection_identifier_only(self):
        assert resolve_application_selection(identifier="app-identifier") == "app-identifier"

    def test_resolve_application_selection_id_only(self):
        assert resolve_application_selection(id=456) == 456

    def test_resolve_application_selection_all_null(self):
        with pytest.raises(UsageError, match="^You must supply one and only one selection value"):
            resolve_application_selection(id_or_identifier=None, id=None, identifier=None)

    @pytest.mark.parametrize(
        "kwargs",
        [
            {"id_or_identifier": "123", "id": 456},
            {"id_or_identifier": "123", "identifier": "app-identifier"},
            {"id": 456, "identifier": "app-identifier"},
        ],
    )
    def test_resolve_application_selection_multiple_args(self, kwargs):
        with pytest.raises(UsageError, match="^You must supply one and only one selection value"):
            resolve_application_selection(**kwargs)


class TestResolveSelection:
    @pytest.mark.parametrize("value", [10, "value", Path("test-test")])
    def test_resolve_selection_single_value(self, value):
        assert resolve_selection(None, value, None) == value

    def test_resolve_selection_no_arguments(self):
        with pytest.raises(UsageError, match="^You must supply one and only one selection value"):
            resolve_selection(option_name="test")

    def test_resolve_selection_all_null(self):
        with pytest.raises(UsageError, match="^You must supply one and only one selection value") as exec_info:
            resolve_selection(None, None, None, option_name="test_option")

        expected_message = "You must supply one and only one selection value (positional test_option or --test-option)"
        assert exec_info.value.message == expected_message

    @pytest.mark.parametrize("value", [10, "value", Path("test-test")])
    def test_resolve_selection_multiple_values(self, value):
        with pytest.raises(UsageError, match="^You must supply one and only one selection value"):
            resolve_selection(value, value)



from jobbergate_cli.text_tools import copy_to_clipboard


def test_copy_to_clipboard__success(mocker):
    """
    Validate that the ``copy_to_clipboard`` function calls ``pyperclip.copy`` successfully and returns ``True``.
    """
    mocked_copy = mocker.patch("pyperclip.copy")
    assert copy_to_clipboard("some text")
    mocked_copy.assert_called_once_with("some text")


def test_copy_to_clipboard__does_not_raise_exception_on_failure(mocker):
    """
    Validate that the ``copy_to_clipboard`` function catches exceptions from ``pyperclip.copy`` and returns ``False``.
    """
    mocked_copy = mocker.patch("pyperclip.copy", side_effect=RuntimeError("BOOM!!!"))
    assert not copy_to_clipboard("some text")
    mocked_copy.assert_called_once_with("some text")



"""
Provide tool functions for working with Application data.
"""

import contextlib
import copy
import io
import pathlib
from dataclasses import dataclass, field
from functools import cached_property
from typing import Any, Dict, List, Union, cast

import yaml
from jobbergate_core.sdk import Apps
from loguru import logger

from jobbergate_cli.constants import (
    JOBBERGATE_APPLICATION_CONFIG,
    JOBBERGATE_APPLICATION_CONFIG_FILE_NAME,
    JOBBERGATE_APPLICATION_MODULE_FILE_NAME,
    JOBBERGATE_APPLICATION_SUPPORTED_FILES,
    FileType,
)
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.render import render_dict, terminal_message
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import (
    ApplicationResponse,
    ContextProtocol,
    JobbergateApplicationConfig,
    JobbergateConfig,
    LocalApplication,
    LocalTemplateFile,
    LocalWorkflowFile,
)
from jobbergate_cli.subapps.applications.application_base import JobbergateApplicationBase
from jobbergate_cli.subapps.applications.questions import inquirer


def load_default_config() -> Dict[str, Any]:
    """
    Load the default config for an application.
    """
    return copy.deepcopy(JOBBERGATE_APPLICATION_CONFIG)


def fetch_application_data_locally(
    application_path: pathlib.Path,
) -> LocalApplication:
    """
    Retrieve an application from a local directory.

    Args:
        application_path: The directory containing the application files.

    Returns:
        A LocalApplication instance containing the application data.
    """
    Abort.require_condition(application_path.is_dir(), f"Application directory {application_path} does not exist")

    config_file_path = application_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
    Abort.require_condition(
        config_file_path.is_file(), f"Application config file {JOBBERGATE_APPLICATION_CONFIG_FILE_NAME} does not exist"
    )

    module_file_path = application_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
    Abort.require_condition(
        module_file_path.is_file(), f"Application module file {JOBBERGATE_APPLICATION_MODULE_FILE_NAME} does not exist"
    )

    template_files_set = set(application_path.rglob("*.j2")) | set(application_path.rglob("*.jinja2"))
    Abort.require_condition(template_files_set, f"No template files found in {application_path}")

    application_config = load_application_config_from_source(config_file_path.read_text())

    supporting_files = application_config.jobbergate_config.supporting_files or []

    template_files = []

    for complete_template_path in template_files_set:
        relative_template_path = complete_template_path.relative_to(application_path)

        if relative_template_path.as_posix() in supporting_files:
            file_type = FileType.SUPPORT
        else:
            file_type = FileType.ENTRYPOINT

        template_files.append(
            LocalTemplateFile(
                filename=relative_template_path.name,
                path=complete_template_path,
                file_type=file_type,
            )
        )

    workflow_file = LocalWorkflowFile(
        filename=module_file_path.name,
        path=module_file_path,
        runtime_config=application_config.jobbergate_config.model_dump(),
    )

    return LocalApplication(
        template_vars=application_config.application_config,
        template_files=template_files,
        workflow_files=[workflow_file],
    )


def fetch_application_data(jg_ctx: ContextProtocol, id_or_identifier: int | str) -> ApplicationResponse:
    """
    Retrieve an application from the API by id or identifier.

    Args:
        jg_ctx: The JobbergateContext. Needed to access the Httpx client with which to make API calls.
        id: The id of the application to fetch.
        identifier: If supplied, look for an application instance with the provided identifier.

    Returns:
        An instance of ApplicationResponse containing the application data.
    """
    stub = f"id={id_or_identifier}" if isinstance(id_or_identifier, int) else f"identifier={id_or_identifier}"
    return cast(
        ApplicationResponse,
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-script-templates/{id_or_identifier}",
            "GET",
            expected_status=200,
            abort_message=f"Couldn't retrieve application {stub} from API",
            response_model_cls=ApplicationResponse,
            support=True,
        ),
    )


@contextlib.contextmanager
def get_upload_files(application_path: pathlib.Path):
    """
    Context manager to build the files parameter.

    Open the supplied file(s) and build a files param appropriate for using
    multi-part file uploads with the client.

    Args:
        application_path: The directory where the application files are located.
    """
    Abort.require_condition(application_path.is_dir(), f"Application directory {application_path} does not exist")

    with contextlib.ExitStack() as stack:
        yield [
            (
                "upload_files",
                (path.name, stack.enter_context(io.open(path, mode="r", newline="")), "text/plain"),
            )
            for path in application_path.rglob("*")
            if path.is_file() and path.suffix in JOBBERGATE_APPLICATION_SUPPORTED_FILES
        ]


def upload_application(
    jg_ctx: ContextProtocol,
    application_path: pathlib.Path,
    id_or_identifier: int | str,
):
    """
    Upload an application given an application path and the application id.

    Args:
        jg_ctx: The JobbergateContext. Needed to access the Httpx client with which to make API calls.
        application_path: The directory where the application files to upload may be found.
        application_id: The id of the application for which to upload data.
        application_identifier: The identifier of the application for which to upload data.
    """
    Abort.require_condition(application_path.is_dir(), f"Application directory {application_path} does not exist")

    config_file_path = application_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
    Abort.require_condition(config_file_path.is_file(), f"Application config file {config_file_path} does not exist")

    module_file_path = application_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
    Abort.require_condition(module_file_path.is_file(), f"Application module file {module_file_path} does not exist")

    template_files_set = set(application_path.rglob("*.j2")) | set(application_path.rglob("*.jinja2"))
    Abort.require_condition(template_files_set, f"No template files found in {application_path}")

    application_config = load_application_config_from_source(config_file_path.read_text())

    logger.debug("Preparing to upload the template configuration")

    make_request(
        jg_ctx.client,
        f"/jobbergate/job-script-templates/{id_or_identifier}",
        "PUT",
        expect_response=False,
        abort_message="Request to upload application configuration was not accepted by the API",
        support=True,
        json={"template_vars": application_config.application_config},
        expected_status=200,
    )

    supporting_files = application_config.jobbergate_config.supporting_files or []

    for complete_template_path in template_files_set:
        relative_template_path = complete_template_path.relative_to(application_path)
        logger.debug(f"Preparing to upload {relative_template_path}")

        with open(complete_template_path, "rb") as template_file:
            if relative_template_path.as_posix() in supporting_files:
                file_type = FileType.SUPPORT
            else:
                file_type = FileType.ENTRYPOINT

            make_request(
                jg_ctx.client,
                f"/jobbergate/job-script-templates/{id_or_identifier}/upload/template/{file_type.value}",
                "PUT",
                expect_response=False,
                abort_message="Request to upload application files was not accepted by the API",
                support=True,
                files={"upload_file": (relative_template_path.name, template_file, "text/plain")},
                expected_status=200,
            )

    logger.debug(f"Preparing to upload {JOBBERGATE_APPLICATION_MODULE_FILE_NAME}")
    with open(module_file_path, "rb") as module_file:
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-script-templates/{id_or_identifier}/upload/workflow",
            "PUT",
            expect_response=False,
            abort_message="Request to upload application module was not accepted by the API",
            support=True,
            files={
                "upload_file": (module_file_path.name, module_file, "text/plain"),
            },
            data={"runtime_config": application_config.jobbergate_config.model_dump_json()},
            expected_status=200,
        )


def save_application_files(
    jg_ctx: ContextProtocol,
    application_data: ApplicationResponse,
    destination_path: pathlib.Path,
) -> List[pathlib.Path]:
    """
    Save the application files from the API response into a local destination.
    """

    logger.debug(f"Saving application files to {destination_path.as_posix()}")
    saved_files: List[pathlib.Path] = []

    if application_data.workflow_files:
        application_config = JobbergateApplicationConfig(
            application_config=application_data.template_vars,
            jobbergate_config=JobbergateConfig(**application_data.workflow_files[0].runtime_config),
        )
        config_path = destination_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
        config_path.parent.mkdir(parents=True, exist_ok=True)
        config_path.write_text(
            yaml.dump(
                application_config.model_dump(
                    mode="json",
                    exclude_none=True,
                    exclude_unset=True,
                ),
                indent=2,
            )
        )
        saved_files.append(config_path)

    for template_file in application_data.template_files:
        template_path = destination_path / "templates" / template_file.filename
        make_request(
            jg_ctx.client,
            template_file.path,
            "GET",
            expected_status=200,
            abort_message=f"Couldn't retrieve template file {template_file.filename} from API",
            save_to_file=template_path,
        )
        saved_files.append(template_path)

    if application_data.workflow_files:
        workflow_path = destination_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
        make_request(
            jg_ctx.client,
            application_data.workflow_files[0].path,
            "GET",
            expected_status=200,
            abort_message="Couldn't retrieve application module file from API",
            save_to_file=workflow_path,
        )
        saved_files.append(workflow_path)

    logger.debug(f"The following files were saved: {list(map(str, saved_files))}")
    return saved_files


def load_application_config_from_source(config_source: str) -> JobbergateApplicationConfig:
    """
    Load the JobbergateApplicationConfig from a text string containing the config as YAML.

    Args:
        config_source: The YAML containing the config

    Returns:
        A JobbergateApplicationConfig instance with the config values
    """
    config_data = yaml.safe_load(config_source)
    config = JobbergateApplicationConfig(**config_data)
    return config


def load_application_from_source(app_source: str) -> type[JobbergateApplicationBase]:
    """
    Load the JobbergateApplication class from a text string containing the source file.

    Creates the module in a temporary file and imports it with importlib.

    Adapted from: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly

    Args:
        app_source: The JobbergateApplication source code to load
    """
    app_locals: Dict[str, Any] = dict()
    exec(app_source, app_locals, app_locals)
    return app_locals["JobbergateApplication"]


@dataclass
class ApplicationRuntime:
    """
    Prepare and execute a Jobbergate application gathering the answers to the questions.

    Args:
        app_data: The application data, can be either an ApplicationResponse or a LocalApplication.
        app_source_code: The source code of the application, often coming from jobbergate.py file.
        supplied_params: The parameters supplied to the application, defaults to an empty dictionary.
        fast_mode: A flag indicating whether the application is in fast mode, defaults to False.
    """

    app_data: Union[ApplicationResponse, LocalApplication]
    app_source_code: str
    sdk: Apps
    supplied_params: Dict[str, Any] = field(default_factory=dict)
    fast_mode: bool = False

    def __post_init__(self) -> None:
        self.answers: Dict[str, Any] = dict()

    @cached_property
    def app_config(self) -> JobbergateApplicationConfig:
        """
        The JobbergateApplicationConfig for the application.
        """
        if not self.app_data.workflow_files:  # make type checker happy
            raise Abort(
                "No workflow file found in application data",
                subject="Invalid application data",
                log_message="No workflow file found in application data",
            )
        try:
            return JobbergateApplicationConfig(
                jobbergate_config=JobbergateConfig(**self.app_data.workflow_files[0].runtime_config),
                application_config=self.app_data.template_vars,
            )
        except Exception as err:
            logger.error("ERR: ", err)
            raise Abort(
                "The application config fetched from the API is not valid",
                subject="Invalid application config",
                support=True,
                log_message="Invalid application config",
                original_error=err,
            )

    @cached_property
    def app_module(self) -> JobbergateApplicationBase:
        """
        The JobbergateApplicationBase for the application.
        """
        try:
            jobbergate_application_class = load_application_from_source(self.app_source_code)
            return jobbergate_application_class(jobbergate_yaml=self.app_config.model_dump(), sdk=self.sdk)
        except Exception as err:
            raise Abort(
                "The application source fetched from the API is not valid",
                subject="Invalid application module",
                support=True,
                log_message="Invalid application module",
                original_error=err,
            )

    def execute_application(self):
        """Execute the jobbergate application python module."""
        try:
            self._gather_answers()
        except Exception as err:
            exception_name = type(err).__name__
            terminal_message(
                "The question workflow failed to execute. Please check the traceback bellow for more information.",
                subject=f"Runtime error on application execution - {exception_name}",
                color="red",
            )
            raise err

        self._update_template_files_information()

    def as_flatten_param_dict(self) -> Dict[str, Any]:
        """Flatten the internal data to support the rendering process."""
        param_dict_flat = {}
        for outer_value in self.app_config.model_dump().values():
            for nest_key, nest_value in outer_value.items():
                param_dict_flat[nest_key] = nest_value
        return param_dict_flat

    def _gather_answers(self):
        """Gather the parameter values by executing the application methods."""
        logger.debug("Gathering answers from the application")
        self.answers.update(self.supplied_params)
        # config should be self.answers ideally
        # but we use self.app_module.jobbergate_config for backward compatibility with v1
        config = self.app_module.jobbergate_config
        config.update(self.supplied_params)

        next_method = "mainflow"

        while next_method is not None:
            method_to_call = getattr(self.app_module, next_method)
            logger.debug(f"Calling method {next_method}")
            try:
                workflow_questions = method_to_call(data=config)
            except NotImplementedError:
                raise Abort(
                    f"""
                    Abstract method not implemented.

                    Please implement {next_method} in your class.",
                    """,
                    subject="Invalid application module",
                )

            prompts = []
            auto_answers = {}

            if workflow_questions is None:
                logger.warning(
                    "Deprecation warning: Application method {} returned None while a list is expected", next_method
                )
                workflow_questions = []

            for question in workflow_questions:
                if question.variablename in self.supplied_params:
                    continue
                elif self.fast_mode and question.default is not None:
                    auto_answers[question.variablename] = question.default
                else:
                    prompts.extend(question.make_prompts())

            workflow_answers = cast(Dict[str, Any], inquirer.prompt(prompts, raise_keyboard_interrupt=True))
            workflow_answers.update(auto_answers)

            logger.debug(f"Answers gathered from {next_method}: {workflow_answers}")

            config.update(workflow_answers)
            self.answers.update(workflow_answers)
            if len(auto_answers) > 0:
                render_dict(auto_answers, title="Default values used")

            next_method = config.pop("nextworkflow", None)

        # Legacy applications change the values at runtime, so we need to update the config
        self.app_config = JobbergateApplicationConfig(
            application_config=dict(self.app_module.application_config),
            jobbergate_config=JobbergateConfig.model_validate(self.app_module.jobbergate_config),
        )
        logger.debug(f"Concluded getting answers: {self.answers}")

    def _update_template_files_information(self):
        """Update the information about the template files if not already present in the configuration."""
        if not self.app_config.jobbergate_config.default_template:
            list_of_entrypoints = [
                i.filename for i in self.app_data.template_files if i.file_type.upper() == "ENTRYPOINT"
            ]
            if len(list_of_entrypoints) != 1:
                raise Abort(
                    f"""
                    Application does not have one entry point, found {len(list_of_entrypoints)}",
                    """,
                    subject="Entry point is unspecified",
                    log_message="Entry point file not specified",
                )
            self.app_config.jobbergate_config.default_template = list_of_entrypoints[0]

        if self.app_config.jobbergate_config.supporting_files is None:
            list_of_supporting_files = [
                i.filename for i in self.app_data.template_files if i.file_type.upper() == "SUPPORT"
            ]
            if list_of_supporting_files:
                self.app_config.jobbergate_config.supporting_files = list_of_supporting_files



import pathlib

from jobbergate_cli.subapps.applications.application_base import JobbergateApplicationBase


class DummyApplication(JobbergateApplicationBase):
    def mainflow(self):
        pass


def test_find_templates(tmp_path):
    application_path = tmp_path / "dummy"
    assert JobbergateApplicationBase.find_templates(application_path) == []

    application_path.mkdir()
    template_root_path = application_path / "templates"
    template_root_path.mkdir()
    file1 = template_root_path / "file1"
    file1.write_text("foo")
    file2 = template_root_path / "file2"
    file2.write_text("bar")
    dir1 = template_root_path / "dir1"
    dir1.mkdir()
    file3 = dir1 / "file3"
    file3.write_text("baz")
    assert JobbergateApplicationBase.find_templates(application_path) == [
        pathlib.Path("templates/dir1/file3"),
        pathlib.Path("templates/file1"),
        pathlib.Path("templates/file2"),
    ]


def test_get_template_files(temp_cd, tmp_path):
    templates_path = tmp_path / "templates"
    templates_path.mkdir()
    template1 = templates_path / "template1.j2"
    template1.write_text("template1")
    template2 = templates_path / "template2.j2"
    template2.write_text("template2")

    subdir_path = templates_path / "subdir"
    subdir_path.mkdir()
    template3 = subdir_path / "template3.j2"
    template3.write_text("template3")

    with temp_cd(tmp_path):
        dummy = DummyApplication(
            dict(
                jobbergate_config=dict(),
                application_config=dict(),
            )
        )
        template_file_paths = dummy.get_template_files()
        assert pathlib.Path("templates/template1.j2") in template_file_paths
        assert pathlib.Path("templates/template2.j2") in template_file_paths
        assert pathlib.Path("templates/subdir/template3.j2") in template_file_paths



import importlib

import pytest
from inquirer import prompt
from inquirer.errors import ValidationError

from jobbergate_cli.subapps.applications.questions import (
    BooleanList,
    Checkbox,
    Confirm,
    Const,
    Directory,
    File,
    Integer,
    List,
    Text,
)


def test_Text__success(dummy_render_class, mocker):
    variablename = "foo"
    question = Text(variablename, "gimme the foo!")
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo="bar")

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    answers = prompt(prompts)
    assert answers["foo"] == "bar"


def test_Integer__success(dummy_render_class, mocker):
    variablename = "foo"
    question = Integer(variablename, "gimme the foo!")
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo=13)

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    answers = prompt(prompts)
    assert answers["foo"] == 13


def test_Integer__zero_as_default():
    """
    The default of zero is an eddy case for inquired.

    Due to implementation details on inquire and the fact that its boolean equivalent
    is false, `0` was not set as default and was not presented on the questions.

    We need to ensure it is.
    """
    question = Integer("foo", "gimme the foo!", default=0)
    prompts = question.make_prompts()

    prompt = prompts.pop()

    assert bool(prompt.default) is True
    assert int(prompt.default) == 0


def test_Integer__fails_with_outside_of_range(dummy_render_class, mocker):
    variablename = "foo"
    question = Integer(variablename, "gimme the foo!", minval=14, maxval=16)
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo=13)

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    with pytest.raises(ValidationError):
        prompt(prompts)

    dummy_render_class.prepared_input = dict(foo=17)

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    with pytest.raises(ValidationError):
        prompt(prompts)


def test_List__success(dummy_render_class, mocker):
    variablename = "foo"
    question = List(variablename, "gimme the foo!", ["a", "b", "c"])
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo="b")

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    answers = prompt(prompts)
    assert answers["foo"] == "b"


def test_Directory__success(tmp_path, dummy_render_class, mocker):
    variablename = "foo"
    question = Directory(variablename, "gimme the foo!", exists=True)
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo=tmp_path)

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    answers = prompt(prompts)
    assert answers["foo"] == tmp_path


def test_Directory__fails_if_directory_does_not_exist(dummy_render_class, mocker):
    variablename = "foo"
    question = Directory(variablename, "gimme the foo!", exists=True)
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo="not/a/real/path")

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    with pytest.raises(ValidationError):
        prompt(prompts)


def test_File__success(tmp_path, dummy_render_class, mocker):
    variablename = "foo"
    question = File(variablename, "gimme the foo!", exists=True)
    prompts = question.make_prompts()

    dummy_file = tmp_path / "dummy"
    dummy_file.write_text("just some dome stuff")
    dummy_render_class.prepared_input = dict(foo=dummy_file.as_posix())

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    answers = prompt(prompts)
    assert answers["foo"] == dummy_file.as_posix()


def test_File__fails_if_file_does_not_exist(dummy_render_class, mocker):
    variablename = "foo"
    question = File(variablename, "gimme the foo!", exists=True)
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo="not/a/real/path")

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    with pytest.raises(ValidationError):
        prompt(prompts)


def test_Checkbox__success(dummy_render_class, mocker):
    variablename = "foo"
    question = Checkbox(variablename, "gimme the foo!", ["a", "b", "c"])
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo=["a", "b"])

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    answers = prompt(prompts)
    assert answers["foo"] == ["a", "b"]


def test_Confirm__success(dummy_render_class, mocker):
    variablename = "foo"
    question = Confirm(variablename, "gimme the foo?")
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict(foo=True)

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    answers = prompt(prompts)
    assert answers["foo"] is True


def test_Const__success(dummy_render_class, mocker):
    variablename = "foo"
    question = Const(variablename, default="bar")
    prompts = question.make_prompts()

    dummy_render_class.prepared_input = dict()

    answers = prompt(prompts, render=dummy_render_class())
    assert answers["foo"] == "bar"


def test_BooleanList__success(dummy_render_class, mocker):
    variablenameTT1 = "fooTT1"
    questionTT1 = Confirm(variablenameTT1, message="gimme the fooTT1!", default=False)

    variablenameT1 = "fooT1"
    questionT1 = BooleanList(variablenameT1, message="gimme the fooT1!", whentrue=[questionTT1], default=False)

    variablenameT2 = "fooT2"
    questionT2 = Confirm(variablenameT2, message="gimme the fooT2!", default=False)

    variablenameFF1 = "fooFF1"
    questionFF1 = Confirm(variablenameFF1, message="gimme the fooFF1!", default=False)

    variablenameF1 = "fooF1"
    questionF1 = BooleanList(variablenameF1, message="gimme the fooF1!", whentrue=[questionFF1], default=False)

    variablename = "foo"
    question = BooleanList(
        variablename,
        message="gimme the foo1!",
        whentrue=[questionT1, questionT2],
        whenfalse=[questionF1],
        default=False,
    )

    prompts = question.make_prompts()

    # We'll answer True to any questions asked to make sure we are ignoring the correct ones
    dummy_render_class.prepared_input = dict(
        fooTT1=True,
        fooT1=True,
        fooT2=True,
        fooFF1=True,
        fooF1=True,
        foo=True,
    )

    mocker.patch.object(importlib.import_module("inquirer.prompt"), "ConsoleRender", new=dummy_render_class)
    answers = dict()
    answers = prompt(prompts, answers=answers, render=dummy_render_class())

    # Only ignored questions should be false
    assert answers == dict(
        fooTT1=True,
        fooT1=True,
        fooT2=True,
        fooFF1=False,
        fooF1=False,
        foo=True,
    )


@pytest.mark.parametrize("parent_answer", [True, False])
def test_BooleanList__same_variable_name(dummy_render_class, parent_answer):
    """Assert that BooleanList works when multiple children have the same variable name."""
    variablename = "child"
    question_a = Text(variablename, message="Question A")
    question_b = Text(variablename, message="Question B")

    question = BooleanList("parent", message="Parent", whentrue=[question_a], whenfalse=[question_b], default=False)
    prompts = question.make_prompts()

    expected_answers = {"parent": parent_answer, variablename: "any-answer"}
    dummy_render_class.prepared_input = expected_answers

    actual_answers = prompt(prompts, answers=expected_answers, render=dummy_render_class())
    assert actual_answers == expected_answers

    expected_ignored_questions = [False, not parent_answer, parent_answer]
    actual_ignored_questions = [q.ignore for q in prompts]
    assert actual_ignored_questions == expected_ignored_questions



from pathlib import Path
from typing import Any, Callable, Dict, Generator, Optional

import httpx
from jobbergate_core import JobbergateAuthHandler
import pytest
import yaml
from typer import Context, Typer
from typer.testing import CliRunner

from jobbergate_cli.constants import JOBBERGATE_APPLICATION_CONFIG_FILE_NAME, JOBBERGATE_APPLICATION_MODULE_FILE_NAME
from jobbergate_cli.context import JobbergateContext
from jobbergate_cli.exceptions import handle_abort, handle_authentication_error
from jobbergate_cli.schemas import IdentityData, JobbergateApplicationConfig, ContextProtocol
from jobbergate_cli.text_tools import dedent


@pytest.fixture
def dummy_domain():
    return "https://dummy.com"


@pytest.fixture
def cli_runner():
    return CliRunner()


@pytest.fixture
def make_test_app(dummy_context):
    def _main_callback(ctx: Context):
        ctx.obj = dummy_context

    def _helper(command_name: str, command_function: Callable):
        main_app = Typer()
        main_app.callback()(_main_callback)
        safe_command = handle_abort(handle_authentication_error(command_function))
        main_app.command(name=command_name)(safe_command)
        return main_app

    return _helper


@pytest.fixture
def dummy_context(mocker, tmp_path, dummy_domain) -> Generator[ContextProtocol, None, None]:
    def dummy_auth(request: httpx.Request) -> httpx.Request:
        request.headers["Authorization"] = "Bearer XXXXXXXX"
        return request

    authentication_handler = JobbergateAuthHandler(cache_directory=Path(tmp_path), login_domain="test-domain")

    context = JobbergateContext()

    with mocker.patch.object(authentication_handler, attribute="acquire_access", return_value=dummy_auth):
        # This is all it takes to replace both cached properties
        context.client = httpx.Client(base_url=dummy_domain)
        context.authentication_handler = authentication_handler
        yield context


@pytest.fixture
def attach_persona(mocker, dummy_context):
    def _helper(
        email: str, client_id: str = "dummy-client", access_token: str = "foo", organization_id: Optional[str] = None
    ):
        identity_data = IdentityData(client_id=client_id, email=email, organization_id=organization_id)
        return mocker.patch.object(
            dummy_context.authentication_handler, attribute="get_identity_data", return_value=identity_data
        )

    return _helper


@pytest.fixture
def seed_clusters(mocker):
    def _helper(*client_ids):
        mocker.patch("jobbergate_cli.subapps.clusters.tools.pull_client_ids_from_api", return_value=client_ids)

    yield _helper


@pytest.fixture
def dummy_application_data():
    return [
        dict(
            id=1,
            name="test-app-1",
            identifier="test-app-1",
            description="Test Application Number 1",
            owner_email="tucker.beck@omnivector.solutions",
            created_at="2022-03-01 17:31:00",
            updated_at="2022-03-01 17:31:00",
            template_vars={"foo": "bar"},
            template_files=[
                {
                    "filename": "test-job-script.py.j2",
                    "parent_id": 1,
                    "created_at": "2022-03-01 17:31:00",
                    "updated_at": "2022-03-01 17:31:00",
                    "file_type": "ENTRYPOINT",
                }
            ],
            workflow_files=[
                {
                    "filename": "jobbergate.py",
                    "parent_id": 1,
                    "created_at": "2022-03-01 17:31:00",
                    "updated_at": "2022-03-01 17:31:00",
                    "runtime_config": {"default_template": "test-job-script.py.j2"},
                },
            ],
        ),
        dict(
            id=2,
            name="test-app-2",
            identifier="test-app-2",
            description="Test Application Number 2",
            owner_email="tucker.beck@omnivector.solutions",
            created_at="2022-03-01 17:31:00",
            updated_at="2022-03-01 17:31:00",
        ),
        dict(
            id=3,
            name="test-app-3",
            identifier="test-app-3",
            description="Test Application Number 3",
            owner_email="tucker.beck@omnivector.solutions",
            created_at="2022-03-01 17:31:00",
            updated_at="2022-03-01 17:31:00",
        ),
    ]


@pytest.fixture
def dummy_job_script_file():
    return {
        "filename": "application.sh",
        "file_type": "ENTRYPOINT",
        "created_at": "2022-03-01 17:31:00",
        "updated_at": "2022-03-01 17:31:00",
    }


@pytest.fixture
def dummy_job_script_data(dummy_application_data, dummy_job_script_file):
    return [
        dict(
            id=1,
            created_at="2022-03-02 22:08:00",
            updated_at="2022-03-02 22:08:00",
            name="script1",
            description="Job Script 1",
            owner_email="tucker@omnivector.solutions",
            application_id=dummy_application_data[0]["id"],
            files=[dict(parent_id=1, **dummy_job_script_file)],
        ),
        dict(
            id=2,
            created_at="2022-03-02 22:08:00",
            updated_at="2022-03-02 22:08:00",
            name="script2",
            description="Job Script 2",
            owner_email="tucker@omnivector.solutions",
            application_id=1,
            files=[dict(parent_id=2, **dummy_job_script_file)],
        ),
        dict(
            id=3,
            created_at="2022-03-02 22:08:00",
            updated_at="2022-03-02 22:08:00",
            name="script3",
            description="Job Script 3",
            owner_email="james@omnivector.solutions",
            application_id=1,
            files=[dict(parent_id=3, **dummy_job_script_file)],
        ),
    ]


@pytest.fixture
def dummy_job_submission_data(dummy_job_script_data):
    return [
        dict(
            id=1,
            created_at="2022-03-02 22:08:00",
            updated_at="2022-03-02 22:08:00",
            name="sub1",
            description="Job Submission 1",
            owner_email="tucker@omnivector.solutions",
            job_script_id=dummy_job_script_data[0]["id"],
            slurm_job_id=13,
            status="CREATED",
        ),
        dict(
            id=1,
            created_at="2022-03-02 22:08:00",
            updated_at="2022-03-02 22:08:00",
            name="sub1",
            description="Job Submission 1",
            owner_email="tucker@omnivector.solutions",
            job_script_id=88,
            slurm_job_id=8888,
            status="CREATED",
        ),
        dict(
            id=3,
            created_at="2022-03-02 22:08:00",
            updated_at="2022-03-02 22:08:00",
            name="sub3",
            description="Job Submission 3",
            owner_email="tucker@omnivector.solutions",
            job_script_id=99,
            slurm_job_id=9999,
            status="CREATED",
        ),
        dict(
            id=4,
            created_at="2022-11-17 11:17:00",
            updated_at="2022-11-17 11:17:00",
            name="sub4",
            description="Job Submission 4",
            owner_email="felipe@omnivector.solutions",
            job_script_id=99,
            slurm_job_id=9999,
            status="REJECTED",
            report_message="Failed to submit job to slurm",
        ),
    ]


@pytest.fixture(scope="module")
def dummy_config_source():
    return dedent(
        """
        jobbergate_config:
          default_template: job-script-template.py.j2
          template_files:
            - job-script-template.py.j2
          supporting_files_output_name:
          supporting_files:
          user_supplied_key: user-supplied-value
        application_config:
          foo: foo
          bar: bar
          baz: baz
        """
    )


@pytest.fixture(scope="module")
def dummy_module_source():
    return dedent(
        """
        from jobbergate_cli.subapps.applications.application_base import JobbergateApplicationBase
        from jobbergate_cli.subapps.applications.questions import Text


        class JobbergateApplication(JobbergateApplicationBase):

            def mainflow(self, data):
                data["nextworkflow"] = "subflow"
                return [Text("foo", message="gimme the foo!"), Text("bar", message="gimme the bar!")]

            def subflow(self, data):
                return [Text("baz", message="gimme the baz!", default="zab")]
        """
    )


@pytest.fixture(scope="module")
def dummy_template_source():
    return dedent(
        """
        #!/bin/python3

        #SBATCH -J dummy_job
        #SBATCH -t 60
        print("I am a very, very dumb job script")
        print(f"foo='{{foo}}'")
        print(f"bar='{{bar}}'")
        print(f"baz='{{baz}}'")
        """
    )


@pytest.fixture
def dummy_jobbergate_application_config(dummy_config_source):
    return JobbergateApplicationConfig(**yaml.safe_load(dummy_config_source))


@pytest.fixture
def dummy_application_dir(tmp_path, dummy_config_source, dummy_module_source, dummy_template_source):
    application_path = tmp_path / "dummy"
    application_path.mkdir()

    module_path = application_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
    module_path.write_text(dummy_module_source)

    config_path = application_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
    config_path.write_text(dummy_config_source)

    template_root_path = application_path / "templates"
    template_root_path.mkdir()

    template_path = template_root_path / "job-script-template.py.j2"
    template_path.write_text(dummy_template_source)

    ignored_path = application_path / "ignored"
    ignored_path.mkdir()

    ignored_file = ignored_path / "ignored.txt"
    ignored_file.write_text("This file should be ignored")

    return application_path


@pytest.fixture
def dummy_render_class():
    class DummyRender:
        prepared_input: Dict[str, Any]

        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

        def render(self, question, answers=None):
            question.answers = answers if answers is not None else dict()

            try:
                ignore = question.ignore(answers)
            except TypeError:
                ignore = question.ignore

            if ignore:
                return question.default

            value = self.prepared_input[question.name]
            question.validate(value)
            return value

    return DummyRender


@pytest.fixture
def dummy_one_page_results():
    return {
        "items": [{"id": 1, "name": "Item 1"}],
        "total": 1,
        "page": 1,
        "size": 50,
        "pages": 1,
    }


@pytest.fixture
def dummy_two_pages_results():
    return [
        {
            "items": [{"id": i + 1, "name": f"Item {i + 1}"} for i in range(50)],
            "total": 80,
            "page": 1,
            "size": 50,
            "pages": 2,
        },
        {
            "items": [{"id": i + 1, "name": f"Item {i + 1}"} for i in range(30)],
            "total": 30,
            "page": 2,
            "size": 50,
            "pages": 2,
        },
    ]



import sys

import pytest
import typer

from jobbergate_cli.compat import add_legacy_compatible_commands


@pytest.fixture
def cleanup_deprecated():
    yield
    for module in list(sys.modules.keys()):
        print("MODULE: ", module)
        if "appform" in module:
            sys.modules.pop(module)
        if "jobberappslib" in module:
            sys.modules.pop(module)
        if "jobbergate_cli.application_base" in module:
            sys.modules.pop(module)


def test_list_all__makes_request_and_renders_results():
    test_app = typer.Typer()
    add_legacy_compatible_commands(test_app)

    registered_command_names = [c.name for c in test_app.registered_commands]

    assert sorted(registered_command_names) == [
        "clone-application",
        "clone-job-script",
        "create-application",
        "create-job-script",
        "create-job-submission",
        "delete-application",
        "delete-job-script",
        "delete-job-submission",
        "download-application",
        "download-job-script",
        "get-application",
        "get-job-script",
        "get-job-submission",
        "list-applications",
        "list-job-scripts",
        "list-job-submissions",
        "render-job-script-locally",
        "show-job-script-files",
        "update-application",
        "update-job-script",
    ]


def test_import_appform_from_jobbergate_cli_warns_and_gives_you_questions_instead_when_in_compatibility_mode(
    tweak_settings, cleanup_deprecated
):
    with tweak_settings(JOBBERGATE_COMPATIBILITY_MODE=True):
        with pytest.warns(DeprecationWarning, match="appform is deprecated"):
            from jobbergate_cli import appform
            from jobbergate_cli.subapps.applications import questions

            assert appform is questions


def test_import_appform_from_jobbergate_cli_raises_an_ImportError_when_not_in_compatibility_mode(cleanup_deprecated):
    with pytest.raises(ImportError):
        from jobbergate_cli import appform  # noqa


def test_import_jobberappslib__from_jobbergate_cli_with_direct_import_of_functions(tweak_settings, cleanup_deprecated):
    with tweak_settings(JOBBERGATE_COMPATIBILITY_MODE=True):
        with pytest.warns(DeprecationWarning, match="Importing jobberappslib from jobbergate_cli is deprecated"):
            from jobbergate_cli.jobberappslib import get_running_jobs as dep_rj
            from jobbergate_cli.subapps.applications.application_helpers import get_running_jobs as new_rj

            assert dep_rj is new_rj


def test_import_jobberappslib_from_jobbergate_cli_raises_ImportError_when_not_in_compatibility_mode(cleanup_deprecated):
    with pytest.raises(ImportError):
        from jobbergate_cli.jobberappslib import get_running_jobs  # noqa


def test_import_application_base_from_jobbergate_cli_warns_and_imports_from_subapp_when_in_compatibility_mode(
    tweak_settings, cleanup_deprecated
):
    with tweak_settings(JOBBERGATE_COMPATIBILITY_MODE=True):
        with pytest.warns(DeprecationWarning, match="application_base from jobbergate_cli is deprecated"):
            from jobbergate_cli.application_base import JobbergateApplicationBase as module_from_base
            from jobbergate_cli.subapps.applications.application_base import (
                JobbergateApplicationBase as module_from_subapp,
            )

            assert module_from_base is module_from_subapp


def test_import_application_base_from_jobbergate_cli_raises_import_error_when_not_in_compatibility_mode(
    cleanup_deprecated,
):
    with pytest.raises(ImportError, match="ApplicationBase has been moved"):
        from jobbergate_cli.application_base import JobbergateApplicationBase  # noqa



import httpx

from jobbergate_cli.constants import PaginationChoices
from jobbergate_cli.schemas import ListResponseEnvelope
from jobbergate_cli.subapps.pagination import handle_pagination


def test_handle_pagination__one_page(respx_mock, dummy_domain, dummy_context, dummy_one_page_results, mocker):
    respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates?page=1").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_one_page_results,
        ),
    )

    mock_inquirer = mocker.patch("jobbergate_cli.subapps.pagination.inquirer.prompt")
    mock_render_paginated = mocker.patch("jobbergate_cli.subapps.pagination.render_paginated_list_results")

    handle_pagination(
        jg_ctx=dummy_context,
        url_path="/jobbergate/job-script-templates",
        params={},
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
    )

    mock_inquirer.assert_not_called()
    mock_render_paginated.assert_called_once_with(
        dummy_context,
        ListResponseEnvelope(**dummy_one_page_results),
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
        value_mappers=None,
    )


def test_handle_pagination__show_next_page(respx_mock, dummy_domain, dummy_context, dummy_two_pages_results, mocker):
    respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates?page=1").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_two_pages_results[0],
        )
    )
    respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates?page=2").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_two_pages_results[1],
        )
    )

    mock_inquirer = mocker.patch("jobbergate_cli.subapps.pagination.inquirer.prompt")
    mock_inquirer.side_effect = [{"navigation": PaginationChoices.NEXT_PAGE}, {"navigation": PaginationChoices.EXIT}]
    mock_render_paginated = mocker.patch("jobbergate_cli.subapps.pagination.render_paginated_list_results")

    handle_pagination(
        jg_ctx=dummy_context,
        url_path="/jobbergate/job-script-templates",
        params={},
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
    )

    assert mock_inquirer.call_count == 2
    assert mock_render_paginated.call_count == 2

    assert mock_render_paginated.call_args_list[0] == mocker.call(
        dummy_context,
        ListResponseEnvelope(**dummy_two_pages_results[0]),
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
        value_mappers=None,
    )
    assert mock_render_paginated.call_args_list[1] == mocker.call(
        dummy_context,
        ListResponseEnvelope(**dummy_two_pages_results[1]),
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
        value_mappers=None,
    )


def test_handle_pagination__show_previous_page(
    respx_mock, dummy_domain, dummy_context, dummy_two_pages_results, mocker
):
    respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates?page=1").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_two_pages_results[0],
        )
    )
    respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates?page=2").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_two_pages_results[1],
        )
    )

    mock_inquirer = mocker.patch("jobbergate_cli.subapps.pagination.inquirer.prompt")
    mock_inquirer.side_effect = [
        {"navigation": PaginationChoices.NEXT_PAGE},
        {"navigation": PaginationChoices.PREVIOUS_PAGE},
        {"navigation": PaginationChoices.EXIT},
    ]
    mock_render_paginated = mocker.patch("jobbergate_cli.subapps.pagination.render_paginated_list_results")

    handle_pagination(
        jg_ctx=dummy_context,
        url_path="/jobbergate/job-script-templates",
        params={},
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
    )

    assert mock_inquirer.call_count == 3
    assert mock_render_paginated.call_count == 3

    assert mock_render_paginated.call_args_list[0] == mocker.call(
        dummy_context,
        ListResponseEnvelope(**dummy_two_pages_results[0]),
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
        value_mappers=None,
    )
    assert mock_render_paginated.call_args_list[1] == mocker.call(
        dummy_context,
        ListResponseEnvelope(**dummy_two_pages_results[1]),
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
        value_mappers=None,
    )
    assert mock_render_paginated.call_args_list[2] == mocker.call(
        dummy_context,
        ListResponseEnvelope(**dummy_two_pages_results[0]),
        title="Applications List",
        style_mapper=None,
        hidden_fields=None,
        value_mappers=None,
    )
